"""
Packages extension
------------------
"""
from __future__ import unicode_literals
import errno
import os
from six.moves.urllib.parse import urljoin
from six import text_type

from lxml.etree import tounicode, SubElement, Element
from lxml.objectify import fromstring

from ..utils.base import ExtensionBase, DataDir
from ..utils.errors import OscError


class Package(ExtensionBase):
    """
    Osc extension to interact with packages
    """
    base_path = "/source"
    new_package_meta_templ = "<package><title/><description/></package>"

    def get_list(self, project, deleted=False):
        """
        Get packages from project

        .. versionadded:: 0.1.7
            Parameter ``deleted``

        :param project: name of project
        :param deleted: Show deleted packages instead
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        params = {"deleted": deleted}
        response = self.osc.request(
            url=urljoin(self.osc.url, "{}/{}".format(self.base_path, project)),
            method="GET",
            params=params
        )

        return self.osc.get_objectified_xml(response)

    def get_meta(self, project, package, blame=False):
        """
        Get package metadata

        .. note:: When ``blame`` annotations are requested no XML object can be
                  returned!

        .. versionchanged:: 0.1.2
            Added parameter blame

        :param project: name of project
        :param package: name of package
        :param blame: Show metadata with change annotations
        :return: Objectified XML element or str
        :rtype: lxml.objectify.ObjectifiedElement or str
        """
        params = {}
        if blame:
            params["view"] = "blame"

        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "{}/{}/{}/_meta".format(self.base_path, project, package)
            ),
            method="GET",
            params=params
        )

        if blame:
            return response.text

        return self.osc.get_objectified_xml(response)

    # pylint: disable=too-many-arguments,protected-access
    def set_meta(self, project, package, title=None, description=None,
                 meta=None):
        """
        Set package metadata

        .. note:: By setting the meta data ``package`` is created, if it does
                  not exist.

        Specify ``title`` and ``description`` to create an XML file with minimum
        content or provide a complete XML string via ``meta``.

        .. note:: If ``title`` or ``description`` is given in combination with
                  ``meta``, the existing values in ``meta`` will be overwritten.

        .. versionadded:: 0.1.2

        :param project: Project name
        :param package: New package name
        :param title: Title for meta
        :param description: Description for meta
        :param meta: New content for meta
        :type meta: str or lxml.objectify.ObjectifiedElement
        :return:
        """
        if isinstance(meta, text_type):
            meta = fromstring(meta)

        if meta is not None:
            meta_xml = meta
        else:
            meta_xml = fromstring(self.new_package_meta_templ)

        if title:
            meta_xml.title._setText(title)
        if description:
            meta_xml.description._setText(description)

        self.osc.request(
            url=urljoin(
                self.osc.url,
                "/".join((self.base_path, project, package, "_meta"))
            ),
            data=tounicode(meta_xml),
            method="PUT"
        )

    def get_files(self, project, package, **params):
        """
        List package files

        :param project: name of project
        :param package: name of package
        :param params: more optional parameters. See:
                       https://build.opensuse.org/apidocs/index#45
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "{}/{}/{}".format(self.base_path, project, package)
            ),
            method="GET",
            params=params
        )

        return self.osc.get_objectified_xml(response)

    # pylint: disable=too-many-arguments
    def get_file(self, project, package, filename, meta=False, rev=None,
                 expand=0):
        """
        Get a source file

        Downloads a specific file from a package and returns a response object
        from which the file contents can be read.

        :param project: name of project
        :param package: name of package
        :param filename: name of file
        :param meta: switch to meta files
        :param rev: Get file from this specific package revision
        :param expand: If the package is linked, it typically only contains a
                       ``_link`` file. In order to allow retrieval of the linked
                       package instead of a 404 error, set this parameter to '1'
        :return: response
        :rtype: requests.Response

        .. versionadded:: 0.1.1
            Parameter rev

        .. versionadded:: 0.1.11
            Parameter expand
        """
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "{}/{}/{}/{}".format(self.base_path, project, package, filename)
            ),
            method="GET",
            stream=True,
            params={'meta': meta, 'rev': rev, 'expand': expand}
        )

        return response

    # pylint: disable=too-many-arguments
    def download_file(self, project, package, filename, destdir, meta=False,
                      overwrite=False, rev=None, expand=False):
        """
        Download a file to directory

        :param project: name of project
        :param package: name of package
        :param filename: name of file
        :param destdir: path of directory
        :param meta: switch to meta files
        :param overwrite: switch to overwrite existing downloaded file
        :param rev: Download file from this specific package revision
        :param expand: If ``True`` and the package is a link, download the file from the linked
                       package
        :return: absolute path to file or ``None``
        :raises OSError: if something goes wrong

        .. versionadded:: 0.1.1
            Parameter rev

        .. versionchanged:: 0.3.3
            Added the parameter ``expand``
        """
        abspath_filename = os.path.abspath(os.path.join(destdir, filename))
        if os.path.isfile(destdir):
            raise OSError(
                errno.EEXIST, "Target directory is a file", destdir
            )
        if not overwrite and os.path.exists(abspath_filename):
            raise OSError(
                errno.EEXIST, "File already exists", abspath_filename
            )
        if not os.path.exists(destdir):
            os.makedirs(destdir)

        response = self.get_file(project, package, filename, meta=meta, rev=rev, expand=expand)

        with open(abspath_filename, "wb") as handle:
            for chunk in response.iter_content(1024):
                handle.write(chunk)

        return abspath_filename

    def push_file(self, project, package, filename, data, comment=None):
        """
        Upload a file to package

        :param project: Name of project
        :param package: Name of package
        :param filename: Name of file
        :param data: content of file
        :type data: str or open file handle
        :param comment: Optional comment to use as commit message

        .. versionchanged:: 0.5.0

           Added an optional ``comment`` argument to be used as the commit message when writing the
           file.
        """
        path = [self.base_path, project, package, filename]

        self.osc.request(
            url=urljoin(self.osc.url, "/".join(path)),
            method="PUT",
            data=data,
            params={"comment": comment}
        )

    def get_attribute(self, project, package, attribute=None):
        """
        Get one attribute of a package

        :param project: name of project
        :param package: name of package
        :param attribute: name of attribute
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        url = urljoin(
            self.osc.url,
            "{}/{}/{}/_attribute".format(
                self.base_path, project, package
            )
        )
        if attribute:
            url = "{}/{}".format(url, attribute)
        response = self.osc.request(
            url=url,
            method="GET"
        )

        return self.osc.get_objectified_xml(response)

    def get_history(self, project, package):
        """
        Get history of package

        :param project: name of project
        :param package: name of package
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "{}/{}/{}/_history".format(self.base_path, project, package)
            ),
            method="GET"
        )

        return self.osc.get_objectified_xml(response)

    def cmd(self, project, package, cmd, **params):
        """
        Get the result of the specified command

        Available commands:

        * diff: Shows the diff of all affected packages.
        * showlinked: List all package instances linking to this one.
        * instantiate: Instantiate a package container, which is available via
          project links only so far.
        * release: Releases sources and binaries of that package.
        * unlock: Unlocks a locked package.
        * branch: Create a source link from a package of an existing project.
        * set_flag: Modify or set a defined flag for package
        * createSpecFileTemplate: Create template for RPM SPEC file.
        * commit: Commits package changes to buildservice.
        * collectbuildenv: Creates _buildenv files based on origin package
          builds.
        * importchannel: Import a kiwi channel file for OBS.

        .. note:: Command ``diff``

            The ``diff`` command returns plain text instaed of XML!

        :param project: name of project
        :param package: name of package
        :param cmd: name of command
        :param params: More command specific parameters. See
                       https://build.opensuse.org/apidocs/index
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement or str
        """
        allowed = [
            'diff', 'showlinked', 'instantiate', 'release', 'unlock', 'branch',
            'set_flag', 'createSpecFileTemplate', 'commit', 'collectbuildenv',
            'importchannel'
        ]

        if cmd not in allowed:
            raise ValueError("Invalid command: '{}'. Use one of: {}".format(
                cmd, ", ".join(allowed)
            ))

        params["cmd"] = cmd
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "{}/{}/{}".format(self.base_path, project, package)
            ),
            method="POST",
            params=params
        )

        if cmd != "diff" or params.get("view", None) == "xml":
            return self.osc.get_objectified_xml(response)

        return response.text

    def checkout(self, project, package, destdir, rev=None, meta=False, expand=False):
        """
        Checkout all files and directories of package

        .. note:: Only by using :py:meth:`checkout` the directory structure is
                  compatible with the ``osc`` command line tool!

        :param project: name of project
        :param package: name of package
        :param destdir: target local directory
        :param rev: Package revision to check out
        :param meta: Checkout meta files instead
        :param expand: If ``True`` and the package is a link, download the file from the linked
                       package
        :return: nothing

        .. versionadded:: 0.1.1

        .. versionchanged:: 0.3.3
            Added the parameter ``expand``
        """
        if not os.path.exists(destdir):
            if not os.path.isdir(destdir):
                os.makedirs(destdir)
            else:
                raise TypeError("Destination {} is a file!".format(destdir))

        oscdir = DataDir(osc=self.osc, path=destdir, project=project,
                         package=package)

        dirlist = self.get_files(project, package, rev=rev, meta=meta, expand=expand)
        for entry in dirlist.findall("entry"):
            self.download_file(
                project=project,
                package=package,
                filename=entry.get("name"),
                destdir=destdir,
                meta=meta,
                overwrite=True,
                rev=rev,
                expand=expand
            )
            os.link(
                os.path.join(destdir, entry.get("name")),
                os.path.join(oscdir.path, entry.get("name"))
            )

    def delete(self, project, package, force=False, comment=None):
        """
        Delete package

        .. versionadded:: 0.1.2

        :param project: Project name
        :param package: Package name
        :param force: Delete package even if pending requests exist
        :param comment: Optional comment
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        params = {'force': force}

        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "/".join((self.base_path, project, package))
            ),
            method="DELETE",
            params=params,
            data=comment
        )

        parsed = self.osc.get_objectified_xml(response)
        if response.status_code == 200 and parsed.get("code") == "ok":
            return True

        return parsed

    def exists(self, project, package, filename=None):
        """
        Check whether package or file in package exists

        .. versionadded:: 0.1.2

        :param project: Project name
        :param package: Package name
        :param filename: Name of file
        :return: ``True``, if package exists, otherwise ``False``
        """
        path = [self.base_path, project, package]
        if filename:
            path.append(filename)
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "/".join(path)
            ),
            method="HEAD",
            raise_for_status=False
        )

        return response.status_code == 200

    # pylint: disable=too-many-locals
    def aggregate(self, src_project, src_package, tgt_project, tgt_package,
                  publish=True, repo_map=None, no_sources=False):
        """
        Aggregate a package to another package

        .. versionadded:: 0.1.2

        .. versionchanged:: 0.2.5

            When creating a new aggregate package, the build flag is always enabled.

        :param src_project: Name of source project
        :param src_package: Name of source package
        :param tgt_project: Name of target project
        :param tgt_package: Name of target package
        :param publish: En-/Disable publishing of aggregated package
        :param repo_map: Optional repository mapping
        :type repo_map: None or dict
        :param no_sources: If ``True``, ignore source packages when copying
                           build results to destination project
        :return:
        """
        # Verify no-op
        if src_project == tgt_project and src_package == tgt_package:
            raise OscError("Source and Target are identical!")

        if not self.exists(src_project, src_package):
            raise OscError("Source package does not exist")

        # Check whether target package exists
        if not self.exists(tgt_project, tgt_package):
            meta_xml = self.get_meta(
                project=src_project,
                package=src_package
            )
            meta_xml.set("name", tgt_package)
            meta_xml.set("project", tgt_project)

            build_elem = meta_xml.find("build")
            if build_elem is None:
                build_elem = SubElement(meta_xml, "build")
            build_elem.clear()
            SubElement(build_elem, "enable")

            if not publish:
                pub_elem = meta_xml.find("publish")
                if not pub_elem:
                    pub_elem = SubElement(meta_xml, "publish")
                pub_elem.clear()
                SubElement(pub_elem, "disable")
            self.set_meta(
                project=tgt_project,
                package=tgt_package,
                meta=meta_xml
            )

        # We do not overwrite an existing aggregate
        if self.exists(tgt_project, tgt_package, "_aggregate"):
            raise OscError("Aggregate already exists.")

        repo_map = repo_map or {}

        # Generate aggregate
        agg_xml = Element("aggregatelist")
        agg = SubElement(agg_xml, "aggregate", project=src_project)
        pkg = SubElement(agg, "package")
        pkg.text = src_package

        if no_sources:
            SubElement(agg, "nosources")

        for src, tgt in repo_map.items():
            SubElement(agg, "repository", target=tgt, source=src)

        self.push_file(
            project=tgt_project,
            package=tgt_package,
            data=tounicode(agg_xml),
            filename="_aggregate"
        )

    def get_comments(self, project, package):
        """
        Get a list of comments for package

        .. versionadded:: 0.1.8
            Use internally :py:class:`osctiny.comments.Comment.get`

        :param project: name of project
        :param package: name of package
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        return self.osc.comments.get(
            obj_type="package",
            ids=(project, package)
        )

    def add_comment(self, project, package, comment, parent_id=None):
        """
        Add a comment to a package

        .. versionchanged:: 0.1.8
            Use internally :py:class:`osctiny.comments.Comment.add`

        :param project: name of project
        :param package: name of package
        :param comment: Comment to be added
        :param parent_id: ID of parent comment. Default: ``None``
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        return self.osc.comments.add(
            obj_type="package",
            ids=(project, package),
            comment=comment,
            parent_id=parent_id
        )



# -*- coding: utf8 -*-
import re

from lxml.objectify import ObjectifiedElement
from requests.exceptions import HTTPError
import responses

from .base import OscTest, CallbackFactory


class TestIssue(OscTest):
    def setUp(self):
        super(TestIssue, self).setUp()

        def callback(headers, params, request):
            status, body = 200, """
            <issue-trackers>
              <issue-tracker>
                <name>boost</name>
                <kind>trac</kind>
                <description>Boost Trac</description>
                <url>https://svn.boost.org/trac/boost/</url>
                <show-url>https://svn.boost.org/trac/boost/ticket/@@@</show-url>
                <regex>boost#(\d+)</regex>
                <label>boost#@@@</label>
                <enable-fetch>false</enable-fetch>
              </issue-tracker>
              <issue-tracker>
                <name>bco</name>
                <kind>bugzilla</kind>
                <description>Clutter Project Bugzilla</description>
                <url>http://bugzilla.clutter-project.org/</url>
                <show-url>http://bugzilla.clutter-project.org/show_bug.cgi?id=@@@</show-url>
                <regex>bco#(\d+)</regex>
                <label>bco#@@@</label>
                <enable-fetch>false</enable-fetch>
              </issue-tracker>
              <issue-tracker>
                <name>bnc</name>
                <kind>bugzilla</kind>
                <description>SUSE Bugzilla</description>
                <url>https://apibugzilla.novell.com/</url>
                <show-url>https://bugzilla.suse.com/show_bug.cgi?id=@@@</show-url>
                <regex>(?:bnc|BNC|bsc|BSC|boo|BOO)\s*[#:]\s*(\d+)</regex>
                <label>bsc#@@@</label>
                <enable-fetch>true</enable-fetch>
              </issue-tracker>
            </issue-trackers>"""
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + '/issue_trackers/?$'),
            callback=CallbackFactory(callback)
        )

    @responses.activate
    def test_get_trackers(self):
        response = self.osc.issues.get_trackers()
        self.assertTrue(isinstance(response, ObjectifiedElement))
        self.assertEqual(response.countchildren(), 3)
        self.assertEqual(
            {x.text for x in response.xpath("issue-tracker/name")},
            {'boost', 'bco', 'bnc'}
        )

    @responses.activate
    def test_get_tracker(self):
        def callback(headers, params, request):
            if re.search("bnc/?$", request.url):
                status, body = 200, """
                <issue-tracker>
                  <name>bnc</name>
                  <kind>bugzilla</kind>
                  <description>SUSE Bugzilla</description>
                  <url>https://apibugzilla.novell.com/</url>
                  <show-url>https://bugzilla.suse.com/show_bug.cgi?id=@@@</show-url>
                  <regex>(?:bnc|BNC|bsc|BSC|boo|BOO)\s*[#:]\s*(\d+)</regex>
                  <label>bsc#@@@</label>
                  <enable-fetch>true</enable-fetch>
                </issue-tracker>"""
            else:
                status, body = 404, """
                <status code="not_found">
                  <summary>Unable to find issue tracker 'bnc2'</summary>
                </status>"""
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + r'/issue_trackers/[^/]+/?$'),
            callback=CallbackFactory(callback)
        )

        with self.subTest("Valid tracker"):
            response = self.osc.issues.get_tracker("bnc")
            self.assertEqual(response.name.text, "bnc")

        with self.subTest("Invalid tracker"):
            self.assertRaises(HTTPError, self.osc.issues.get_tracker, "nemo")

    @responses.activate
    def test_get(self):
        def callback(headers, params, request):
            if params.get("force_update", ["0"]) == ["1"]:
                status, body = 200, u"""
                <issue>
                  <created_at>2020-01-04 14:12:00 UTC</created_at>
                  <name>1160086</name>
                  <tracker>bnc</tracker>
                  <label>bsc#1160086</label>
                  <url>https://bugzilla.suse.com/show_bug.cgi?id=1160086</url>
                  <state>OPEN</state>
                  <summary>føø bar</summary>
                  <owner>
                    <login>nemo</login>
                    <email>nemo@suse.com</email>
                    <realname>Caþtæn Nemo</realname>
                  </owner>
                </issue>"""
            else:
                status, body = 200, """
                <issue>
                  <created_at>2020-01-04 14:12:00 UTC</created_at>
                  <name>1160086</name>
                  <tracker>bnc</tracker>
                  <label>bsc#1160086</label>
                  <url>https://bugzilla.suse.com/show_bug.cgi?id=1160086</url>
                </issue>"""
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url +
                           r'/issue_trackers/bnc/issues/1160086/?.*'),
            callback=CallbackFactory(callback)
        )

        with self.subTest("Manual force update"):
            response = self.osc.issues.get("bnc", 1160086, True)
            self.assertTrue(hasattr(response, "summary"))
            self.assertEqual(len(responses.calls), 2)

        with self.subTest("Manual force update"):
            response = self.osc.issues.get("bnc", 1160086, False)
            self.assertTrue(hasattr(response, "summary"))
            # to whom it may concern: `responses.calls` does not get reset
            # between sub-tests
            self.assertEqual(len(responses.calls), 4)



import re
from urllib.parse import urlparse, parse_qs

from requests.exceptions import HTTPError
import responses

from .base import OscTest, CallbackFactory


class BuildTest(OscTest):

    @responses.activate
    def test_get(self):
        def callback(headers, params, request):
            status = 500
            body = ""
            parsed = urlparse(request.url)
            params.update(parse_qs(parsed.query))

            if not params:
                status = 200
                body = """
                <resultlist state="2c371787daedadb61dfd5fd8411ac6c2">
                  <result project="Some:Project" repository="SLE_15" 
                          arch="x86_64" code="published" state="published">
                    <status package="mypackage" code="disabled" />
                    <status package="anotherpackage" code="disabled" />
                  </result>
                  <result project="Some:Project" repository="SLE_12_SP3" 
                          arch="x86_64" code="published" state="published">
                    <status package="mypackage" code="succeeded" />
                    <status package="anotherpackage" code="disabled" />
                  </result>
                  <result project="Some:Project" repository="SLE_12_SP2" 
                          arch="x86_64" code="published" state="published">
                    <status package="mypackage" code="disabled" />
                    <status package="anotherpackage" code="disabled" />
                  </result>
                </resultlist>
                """
            elif params.get("package", []):
                if params["package"][0] in ["mypackage", "anotherpackage"]:
                    status = 200
                    body = """
                    <resultlist state="110f481609293ee149e28bbaced3b1b9">
                      <result project="Some:Project"  repository="SLE_15" 
                              arch="x86_64" code="published" state="published">
                        <status package="{pkg}" code="disabled" />
                      </result>
                      <result project="Some:Project" repository="SLE_12_SP3" 
                              arch="x86_64" code="published" state="published">
                        <status package="{pkg}" code="succeeded" />
                      </result>
                      <result project="Some:Project" repository="SLE_12_SP2" 
                              arch="x86_64" code="published" state="published">
                        <status package="{pkg}" code="disabled" />
                      </result>
                    </resultlist>
                    """.format(pkg=params["package"][0])
                else:
                    status = 404
                    body = """
                    <status code="404">
                      <summary>unknown package '{pkg}'</summary>
                      <details>404 unknown package '{pkg}'</details>
                    </status>
                    """.format(pkg=params["package"][0])

            return status, headers, body

        project = "Some:Project"
        package = "mypackage"
        self.mock_request(
            method=responses.GET,
            url=re.compile(
                self.osc.url + "/build/{}/_result".format(project)
            ),
            callback=CallbackFactory(callback)
        )

        with self.subTest("all"):
            response = self.osc.build.get(project=project)
            self.assertEqual(len(response.findall("result/status")), 6)

        with self.subTest("package: mypackage"):
            response = self.osc.build.get(project=project, package=package)
            self.assertEqual(len(response.findall("result/status")), 3)

        with self.subTest("package: anotherpackage"):
            response = self.osc.build.get(project=project,
                                          package="anotherpackage")
            self.assertEqual(len(response.findall("result/status")), 3)

        with self.subTest("package: unknown"):
            self.assertRaises(
                HTTPError,
                self.osc.build.get, project=project, package="xyz123"
            )



from unittest import mock

# Absolute import needed for mocking ;)
from ..utils.base import DataDir
from .base import OscTest


class DataDirTest(OscTest):

    @mock.patch("os.path.isdir", return_value=True)
    @mock.patch("osctiny.utils.base.open", create=True, new_callable=mock.mock_open)
    def test_datadir_for_project_1(self, open_mock, isdir_mock):
        # Data dir exists and overwrite=False
        oscdir = DataDir(self.osc, "/some/path", "Some:Project")
        self.assertEqual(open_mock.call_count, 0)
        self.assertEqual(isdir_mock.call_count, 1)

    @mock.patch("os.makedirs")
    @mock.patch("os.path.isdir", return_value=False)
    @mock.patch("osctiny.utils.base.open", create=True, new_callable=mock.mock_open)
    def test_datadir_for_project_2(self, open_mock, isdir_mock, makedirs_mock):
        # Data dir does not exist
        oscdir = DataDir(self.osc, "/some/path", "Some:Project")
        self.assertEqual(open_mock.call_count, 3)
        self.assertEqual(len(open_mock.mock_calls), 12)
        self.assertEqual(isdir_mock.call_count, 1)

    @mock.patch("os.makedirs")
    @mock.patch("os.path.isdir", return_value=True)
    @mock.patch("osctiny.utils.base.open", create=True, new_callable=mock.mock_open)
    def test_datadir_for_project_3(self, open_mock, isdir_mock, makedirs_mock):
        # Data dir exists and overwrite=True
        oscdir = DataDir(self.osc, "/some/path", "Some:Project", overwrite=True)
        self.assertEqual(open_mock.call_count, 3)
        self.assertEqual(len(open_mock.mock_calls), 12)
        self.assertEqual(isdir_mock.call_count, 1)



import re
from urllib.parse import urlparse, parse_qs
import responses

from .base import OscTest, CallbackFactory


def callback(headers, params, request):
    status = 500
    body = ""
    parsed = urlparse(request.url)
    params.update(parse_qs(parsed.query))

    if "project" in parsed.path:
        status = 200
        body = """
<collection matches="2">
    <project name="SUSE:Maintenance:96" kind="maintenance_incident">
      <title/>
      <description/>
      <person userid="krahmer" role="bugowner"/>
      <person userid="oertel" role="maintainer"/>
      <group groupid="maintenance-team" role="maintainer"/>
      <group groupid="autobuild-team" role="reader"/>
      <group groupid="maintenance-team" role="reviewer"/>
      <lock> 
        <enable/>
      </lock>
      <build>
        <disable/>
      </build>
      <publish>
        <disable/>
      </publish>
      <debuginfo>
        <enable/>
      </debuginfo>
      <repository name="SUSE_Updates_SLE-SERVER_12_x86_64">
        <releasetarget project="SUSE:Updates:SLE-SERVER:12:x86_64" 
          repository="update"/>
        <path project="SUSE:Updates:SLE-SERVER:12:x86_64" repository="update"/>
        <arch>x86_64</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SERVER_12_s390x">
        <releasetarget project="SUSE:Updates:SLE-SERVER:12:s390x" 
          repository="update"/>
        <path project="SUSE:Updates:SLE-SERVER:12:s390x" repository="update"/>
        <arch>s390x</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SERVER_12_ppc64le">
        <releasetarget project="SUSE:Updates:SLE-SERVER:12:ppc64le" 
          repository="update"/>
        <path project="SUSE:Updates:SLE-SERVER:12:ppc64le" repository="update"/>
        <arch>ppc64le</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-DESKTOP_12_x86_64">
        <releasetarget project="SUSE:Updates:SLE-DESKTOP:12:x86_64" 
          repository="update"/>
        <path project="SUSE:Updates:SLE-DESKTOP:12:x86_64" repository="update"/>
        <arch>x86_64</arch>
      </repository>
      <repository name="SUSE_SLE-12_Update">
        <releasetarget project="SUSE:SLE-12:Update" repository="standard"/>
        <path project="SUSE:SLE-12:Update" repository="standard"/>
        <arch>i586</arch>
        <arch>x86_64</arch>
        <arch>s390</arch>
        <arch>s390x</arch>
        <arch>ppc64le</arch>
        <arch>aarch64</arch>
      </repository>
    </project>
    <project name="SUSE:Maintenance:9691" kind="maintenance_incident">
      <title/>
      <description/>
      <person userid="gboiko" role="bugowner"/>
      <person userid="oertel" role="maintainer"/>
      <group groupid="maintenance-team" role="maintainer"/>
      <group groupid="autobuild-team" role="reviewer"/>
      <group groupid="legal-auto" role="reviewer"/>
      <group groupid="maintenance-team" role="reviewer"/>
      <build>
        <disable/>
      </build>
      <publish>
        <disable/>
      </publish>
      <debuginfo>
        <enable/>
      </debuginfo>
      <repository name="SUSE_Updates_SLE-SERVER_12-SP4_x86_64">
        <releasetarget project="SUSE:Updates:SLE-SERVER:12-SP4:x86_64" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SERVER:12-SP4:x86_64" 
          repository="update"/>
        <arch>x86_64</arch>
        <arch>i586</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SERVER_12-SP4_s390x">
        <releasetarget project="SUSE:Updates:SLE-SERVER:12-SP4:s390x" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SERVER:12-SP4:s390x" 
          repository="update"/>
        <arch>s390x</arch>
        <arch>s390</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SERVER_12-SP4_ppc64le">
        <releasetarget project="SUSE:Updates:SLE-SERVER:12-SP4:ppc64le" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SERVER:12-SP4:ppc64le" 
          repository="update"/>
        <arch>ppc64le</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SERVER_12-SP4_aarch64">
        <releasetarget project="SUSE:Updates:SLE-SERVER:12-SP4:aarch64" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SERVER:12-SP4:aarch64" 
          repository="update"/>
        <arch>aarch64</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SDK_12-SP4_x86_64">
        <releasetarget project="SUSE:Updates:SLE-SDK:12-SP4:x86_64" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SDK:12-SP4:x86_64" repository="update"/>
        <arch>x86_64</arch>
        <arch>i586</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SDK_12-SP4_s390x">
         <releasetarget project="SUSE:Updates:SLE-SDK:12-SP4:s390x" 
           repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SDK:12-SP4:s390x" repository="update"/>
        <arch>s390x</arch>
        <arch>s390</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SDK_12-SP4_ppc64le">
        <releasetarget project="SUSE:Updates:SLE-SDK:12-SP4:ppc64le" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SDK:12-SP4:ppc64le" repository="update"/>
        <arch>ppc64le</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-SDK_12-SP4_aarch64">
        <releasetarget project="SUSE:Updates:SLE-SDK:12-SP4:aarch64" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-SDK:12-SP4:aarch64" repository="update"/>
        <arch>aarch64</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-HA_12-SP4_x86_64">
        <releasetarget project="SUSE:Updates:SLE-HA:12-SP4:x86_64" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-HA:12-SP4:x86_64" repository="update"/>
        <arch>x86_64</arch>
        <arch>i586</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-HA_12-SP4_s390x">
        <releasetarget project="SUSE:Updates:SLE-HA:12-SP4:s390x" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-HA:12-SP4:s390x" repository="update"/>
        <arch>s390x</arch>
        <arch>s390</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-HA_12-SP4_ppc64le">
        <releasetarget project="SUSE:Updates:SLE-HA:12-SP4:ppc64le" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-HA:12-SP4:ppc64le" repository="update"/>
        <arch>ppc64le</arch>
      </repository>
      <repository name="SUSE_Updates_SLE-DESKTOP_12-SP4_x86_64">
        <releasetarget project="SUSE:Updates:SLE-DESKTOP:12-SP4:x86_64" 
          repository="update" trigger="maintenance"/>
        <path project="SUSE:Updates:SLE-DESKTOP:12-SP4:x86_64" 
          repository="update"/>
        <arch>x86_64</arch>
        <arch>i586</arch>
      </repository>
      <repository name="SUSE_SLE-12-SP4_Update">
        <releasetarget project="SUSE:SLE-12-SP4:Update" repository="standard" 
          trigger="maintenance"/>
        <path project="SUSE:SLE-12-SP4:Update" repository="standard"/>
        <arch>i586</arch>
        <arch>x86_64</arch>
        <arch>s390</arch>
        <arch>s390</arch>
        <arch>s390x</arch>
        <arch>ppc64le</arch>
        <arch>aarch64</arch>
      </repository>
    </project>
</collection>
        """
    elif "package" in parsed.path:
        status = 200
        body = """
<collection matches="6">
<package name="python-django" project="SUSE:SLE-11-SP3:GA:Products:Test">
  <title>A high-level Python Web framework</title>
  <description>Django is a high-level Python Web framework that encourages rapid
  development and clean, pragmatic design.</description>
  <person userid="cloud_bugs" role="bugowner"/>
</package>
<package name="python-django" project="Devel:Cloud:1.0">
  <title>The Web framework for perfectionists with deadlines</title>
  <description>Django is a high-level Python Web framework that encourages rapid
  development and clean, pragmatic design.

Developed by a fast-moving online-news operation, Django was designed to handle 
two challenges: the intensive deadlines of a newsroom and the stringent 
requirements of the experienced Web developers who wrote it. It lets you build 
high-performing, elegant Web applications quickly.

Django focuses on automating as much as possible and adhering to the DRY 
principle.</description>
  <url>http://www.djangoproject.com/</url>
</package>
<package name="python-django" project="Devel:Cloud:2.0">
  <title/>
  <description/>
</package>
<package name="python-django" project="SUSE:SLE-11-SP3:Update">
  <title>A high-level Python Web framework</title>
  <description>Django is a high-level Python Web framework that encourages rapid
  development and clean, pragmatic design.</description>
  <url>http://www.djangoproject.com/</url>
</package>
<package name="python-django" project="SUSE:SLE-11-SP2:Update">
  <title>A high-level Python Web framework</title>
  <description>Django is a high-level Python Web framework that encourages rapid
  development and clean, pragmatic design.</description>
  <releasename>python-django</releasename>
  <url>http://www.djangoproject.com/</url>
</package>
<package name="python-django" project="Devel:Cloud:Shared:11-SP3:Update">
  <title/>
  <description/>
</package>
</collection>
        """
    elif "request" in parsed.path:
        status = 200
        body = """
<collection matches="">
  <request id="167315" creator="aaa">
    <action type="submit">
      <source 
        project="home:aaa:branches:SUSE:SLE-12-SP3:Update:Products:SES5:Update" 
        package="python-Django" rev="dbc59e44f0ea89fb37a081328f7d9b95"/>
      <target project="Devel:Storage:5.0" package="python-Django"/>
      <acceptinfo rev="2" srcmd5="9e6269ae5b5933a3593c413e8816ac7b" 
      osrcmd5="015b33968bcb0deb50d5b702826ddf6a"/>
    </action>
    <state name="accepted" who="aaa" when="2018-06-21T11:07:41">
      <comment>Just me fixing up the changelog</comment>
    </state>
    <description>- Prevented spoofing is_safe_url() with basic auth (bsc#967999,
      CVE-2016-2512)</description>
  </request>
  <request id="167473" creator="aaa">
    <action type="submit">
      <source 
      project="home:aaa:branches:SUSE:SLE-12-SP3:Update:Products:SES5:Update" 
      package="python-Django" rev="691eedaf40acd37e4e682a1668123c85"/>
      <target project="Devel:Storage:5.0" package="python-Django"/>
      <acceptinfo rev="3" srcmd5="8918eba54a899896eaa623a3c2ee8452" 
        osrcmd5="9e6269ae5b5933a3593c413e8816ac7b"/>
    </action>
    <state name="accepted" who="bbb" when="2018-06-25T11:21:20">
      <comment/>
    </state>
    <description>- Fixed catastrophic backtracking in urlize and urlizetrunc 
    template filters
    (bsc#1083304, CVE-2018-7536)
    * Added CVE-2018-7536.patch
  - Fixed catastrophic backtracking in django.utils.text.Truncator (bsc#1083305,
    CVE-2018-7537)
    * Added CVE-2018-7537.patch</description>
  </request>
  <request id="167474" creator="aaa">
    <action type="submit">
      <source 
      project="home:aaa:branches:SUSE:SLE-12-SP2:Update:Products:SES4:Update" 
      package="python-Django" rev="f4f53a48efc315e7e339d84b00f8f74e"/>
      <target project="Devel:Storage:4.0" package="python-Django"/>
      <acceptinfo rev="3" srcmd5="8918eba54a899896eaa623a3c2ee8452" 
        osrcmd5="9e6269ae5b5933a3593c413e8816ac7b"/>
    </action>
    <state name="accepted" who="ccc" when="2018-06-25T11:22:10">
      <comment/>
    </state>
    <description>- Fixed catastrophic backtracking in urlize and urlizetrunc 
    template filters
    (bsc#1083304, CVE-2018-7536)
    * Added CVE-2018-7536.patch
  - Fixed catastrophic backtracking in django.utils.text.Truncator (bsc#1083305,
    CVE-2018-7537)
    * Added CVE-2018-7537.patch</description>
  </request>
</collection>

        """

    headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
    return status, headers, body


class TestSearch(OscTest):
    def setUp(self):
        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + '/search.*'),
            callback=CallbackFactory(callback)
        )

    @responses.activate
    def test_search(self):
        with self.subTest("Project"):
            response = self.osc.search.project(
                "@kind='maintenance_incident' and "
                "starts-with(@name,'SUSE:Maintenance')"
            )
            self.assertEqual(response.tag, "collection")
            self.assertEqual(response.countchildren(), 2)
            self.assertEqual(
                sorted([x.get("name", None) for x in response.getchildren()]),
                sorted(["SUSE:Maintenance:96", "SUSE:Maintenance:9691"])
            )

        with self.subTest("Package"):
            response = self.osc.search.package(
                "@name='python-django'"
            )
            self.assertEqual(response.tag, "collection")
            self.assertEqual(response.countchildren(), 6)
            self.assertEqual(
                sorted([x.get("project", None) for x in response.getchildren()]),
                sorted([
                    'Devel:Cloud:2.0', 'SUSE:SLE-11-SP3:GA:Products:Test',
                    'Devel:Cloud:1.0', 'SUSE:SLE-11-SP3:Update',
                    'SUSE:SLE-11-SP2:Update', 'Devel:Cloud:Shared:11-SP3:Update'
                ])
            )

        with self.subTest("Request"):
            response = self.osc.search.request(
                "target[@package='python-django']"
            )
            self.assertEqual(response.tag, "collection")
            self.assertEqual(response.countchildren(), 3)
            self.assertEqual(
                sorted(
                    [x.get("who", "zzz")
                     for x in response.xpath("request/state")]
                ),
                ["aaa", "bbb", "ccc"]
            )



# -*- coding: utf8 -*-
import re

from requests.exceptions import HTTPError
import responses

from .base import OscTest, CallbackFactory


class TestDistribution(OscTest):
    @responses.activate
    def test_get_list(self):
        def _base(include_remotes):
            response = self.osc.distributions.get_list(include_remotes=include_remotes)
            self.assertEqual({int(d.get("id")) for d in response.distribution},
                             {14495, 14498, 14501})

        def dist_list_callback(headers, params, request):
            status, body = 200, """
            <distributions>
              <distribution vendor="openSUSE" version="Tumbleweed" id="14495">
                <name>openSUSE Tumbleweed</name>
                <project>openSUSE:Factory</project>
                <reponame>openSUSE_Tumbleweed</reponame>
                <repository>snapshot</repository>
                <link>http://www.opensuse.org/</link>
                <architecture>i586</architecture>
                <architecture>x86_64</architecture>
              </distribution>
              <distribution vendor="openSUSE" version="15.2" id="14498">
                <name>openSUSE Leap 15.2</name>
                <project>openSUSE:Leap:15.2</project>
                <reponame>openSUSE_Leap_15.2</reponame>
                <repository>standard</repository>
                <link>http://www.opensuse.org/</link>
                <architecture>x86_64</architecture>
              </distribution>
              <distribution vendor="openSUSE" version="15.1" id="14501">
                <name>openSUSE Leap 15.1</name>
                <project>openSUSE:Leap:15.1</project>
                <reponame>openSUSE_Leap_15.1</reponame>
                <repository>standard</repository>
                <link>http://www.opensuse.org/</link>
                <architecture>x86_64</architecture>
              </distribution>
            </distributions>"""
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + r'/distributions(/include_remotes)?/?$'),
            callback=CallbackFactory(dist_list_callback)
        )

        with self.subTest("include remotes"):
            _base(True)

        with self.subTest("do not include remotes"):
            _base(False)

    @responses.activate
    def test_get(self):
        def dist_callback(headers, params, request):
            if re.search("14498/?$", request.url):
                status, body = 200, """<?xml version="1.0" encoding="UTF-8"?>
                <hash>
                  <id type="integer">14498</id>
                  <vendor>openSUSE</vendor>
                  <version>15.2</version>
                  <name>openSUSE Leap 15.2</name>
                  <project>openSUSE:Leap:15.2</project>
                  <reponame>openSUSE_Leap_15.2</reponame>
                  <repository>standard</repository>
                  <link>http://www.opensuse.org/</link>
                  <architectures type="array">
                    <architecture>x86_64</architecture>
                  </architectures>
                </hash>"""
            else:
                status, body = 404, """
                <status code="not_found">
                  <summary>Couldn't find Distribution with 'id'=******</summary>
                </status>"""
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + r'/distributions/\d+/?$'),
            callback=CallbackFactory(dist_callback)
        )

        with self.subTest("existing distro"):
            response = self.osc.distributions.get(14498)
            self.assertEqual(response.id, 14498)
            self.assertEqual(response.vendor, "openSUSE")
            self.assertEqual(response.version, 15.2)

        with self.subTest("unknown distro"):
            self.assertRaises(HTTPError, self.osc.distributions.get, 99999)



# -*- coding: utf-8 -*-
from base64 import b64encode
from bz2 import compress
from unittest import TestCase, mock
from datetime import datetime
from io import StringIO
import os
from pathlib import Path
import sys
from tempfile import mkstemp
from types import GeneratorType

from dateutil.parser import parse
from pytz import _UTC, timezone

from ..utils.changelog import ChangeLog, Entry
from ..utils.conf import get_config_path, get_credentials
from ..utils.mapping import Mappable

sys.path.append(os.path.dirname(__file__))

SAMPLE_CHANGES = """
-------------------------------------------------------------------
Tue Dec  3 10:38:41 UTC 2019 - Andreas Hasenkopf <ahasenkopf@suse.com>

- Version 0.1.10

-------------------------------------------------------------------
Tue Dec  3 10:13:08 UTC 2019 - Andreas Hasenkopf <ahasenkopf@suse.com>

- New package osc-tiny (version 0.1.9)
"""

SAMPLE_CHANGES_2 = """
-------------------------------------------------------------------
Mon Aug  5 06:51:32 CST 2019 - Marco Strigl <marco.strigl@suse.com>

- 0.165.4 (boo#1144211)
   * allow optional fork when creating a maintenance request
   * fix RPMError fallback
   * fix local caching for all package formats
   * fix appname for trusted cert store
   * osc -h does not break anymore when using plugins 

-------------------------------------------------------------------
Wed Jul 24 13:18:01 UTC 2019 - Marco Strigl <marco.strigl@suse.com>

- 0.165.3 (boo#1142662)
    * switch to difflib.diff_bytes and sys.stdout.buffer.write for diffing.
      This will fix all decoding issues with osc diff, osc ci and osc rq -d
    * fix osc ls -lb handling empty size and mtime
    * removed decoding on osc api command.
    * fixed broken TLS certificate handling (boo#1142518, CVE-2019-3685) 

-------------------------------------------------------------------
Fri May 12 00:00:00 CEST 2006 - poeml@suse.de

- don't use --record-rpm option on setup.py, only SUSE has it
- define py_sitelib macro

-------------------------------------------------------------------
Wed May 10 00:00:00 CEST 2006 - poeml@suse.de

- created package (version 0.2)


"""

SAMPLE_CHANGES_3 = """
Tue Dec  3 10:38:41 UTC 2019 - Andreas Hasenkopf <ahasenkopf@suse.com>

- Version 0.1.10

-------------------------------------------------------------------
Tue Dec  3 10:13:08 UTC 2019 - Andreas Hasenkopf <ahasenkopf@suse.com>

- New package osc-tiny (version 0.1.9)
"""


class TestMappable(TestCase):
    def test(self):
        m = Mappable(a="a", b="b")
        m["c"] = "c"

        for key in ('a', 'b', 'c'):
            with self.subTest(f"get {key}"):
                self.assertEqual(key, m.get(key))

        with self.subTest("Default"):
            self.assertEqual("føø", m.get("d", "føø"))


class TestEntry(TestCase):
    def test_timestamp(self):
        cet = timezone("Europe/Berlin")

        with self.subTest("Initialization"):
            entry = Entry(timestamp=datetime(2019, 1, 1, tzinfo=cet))
            self.assertEqual(entry.timestamp.tzinfo, cet)
            self.assertIn("UTC", entry.formatted_timestamp)

        with self.subTest("Assignment"):
            entry = Entry()
            self.assertIsNotNone(entry.timestamp)
            entry.timestamp = datetime(2019, 1, 1, tzinfo=cet)
            self.assertIn("UTC", entry.formatted_timestamp)

        with self.subTest("Naive datetime"):
            self.assertRaises(ValueError, Entry, timestamp=datetime(2019, 1, 1))


class TestChangeLog(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TestChangeLog, cls).setUpClass()
        cls.utc = _UTC()

    @staticmethod
    def prep_mock_open(omock):
        """
        Make a mocked file handle iterable

        See:

        * https://stackoverflow.com/a/24779923
        * https://docs.python.org/3.8/library/unittest.mock.html#mock-open
        """
        omock.return_value.__iter__ = lambda self: self
        omock.return_value.__next__ = lambda self: next(iter(self.readline, ''))

    def test_pattern(self):
        headers = [
            "Thu Dec  5 08:57:27 UTC 2019 - Adrian Schröter <adrian@suse.de>",
            "Wed Apr 24 09:37:32 CEST 2019 - schubi@suse.de",
            "Tue Dec  3 10:13:08 UTC 2019 - Andreas Hasenkopf "
            "<ahasenkopf@suse.com>"
        ]

        with self.subTest("Header pattern"):
            for line in headers:
                match = ChangeLog.patterns["header"].match(line)
                self.assertIsNotNone(match)

                timestamp = parse(match.group("timestamp"))
                self.assertIsInstance(timestamp, datetime)

    def test_parse_non_generative(self):
        with mock.patch("osctiny.utils.changelog.open",
                        mock.mock_open(read_data=SAMPLE_CHANGES),
                        create=True) as omock:
            self.prep_mock_open(omock)
            cl = ChangeLog.parse("/who/cares/test.changes", generative=False)

        self.assertIsInstance(cl.entries, list)
        self.assertEqual(len(cl.entries), 2)
        self.assertEqual(cl.entries[0].packager,
                         "Andreas Hasenkopf <ahasenkopf@suse.com>")
        self.assertEqual({x.content for x in cl.entries},
                         {
                             "- New package osc-tiny (version 0.1.9)",
                             "- Version 0.1.10"
                         })

    def test_parse_stringio(self):
        buffer = StringIO(SAMPLE_CHANGES)
        cl = ChangeLog.parse(buffer, generative=False)

        self.assertIsInstance(cl.entries, list)
        self.assertEqual(len(cl.entries), 2)

    def test_parse_path(self):
        _, path = mkstemp()
        with open(path, "w") as handle:
            handle.write(SAMPLE_CHANGES)

        try:
            with self.subTest("generative"):
                cl = ChangeLog.parse(path, generative=True)

                self.assertNotIsInstance(cl.entries, list)
                self.assertEqual(len(list(cl.entries)), 2)

            with self.subTest("non-generative"):
                cl = ChangeLog.parse(path, generative=False)

                self.assertIsInstance(cl.entries, list)
                self.assertEqual(len(cl.entries), 2)
        finally:
            os.remove(path)

    def test_parse_generative(self):
        with mock.patch("osctiny.utils.changelog.open",
                        mock.mock_open(read_data=SAMPLE_CHANGES),
                        create=True) as omock:
            self.prep_mock_open(omock)
            cl = ChangeLog.parse("/who/cares/test.changes", generative=True)

            self.assertIsInstance(cl.entries, GeneratorType)

            counter = 0
            # Notice: This loop must be within the mocked context!
            for _ in cl.entries:
                counter += 1

            self.assertEqual(counter, 2)

    def test_write(self):
        with mock.patch("osctiny.utils.changelog.open",
                        mock.mock_open(read_data=""),
                        create=True) as omock:
            self.prep_mock_open(omock)
            cl = ChangeLog()
            cl.entries = [
                Entry(
                    packager="Andreas Hasenkopf <ahasenkopf@suse.com>",
                    content="Føø Bar",
                    timestamp=datetime(2019, 1, 1, 0, 0, 0, tzinfo=self.utc)
                ),
                Entry(
                    packager="Andreas Pritschet <apritschet@suse.com>",
                    content="Hellø Wørld",
                    timestamp=datetime(2018, 1, 1, 0, 0, 0, tzinfo=self.utc)
                ),
                Entry(
                    packager="Andreas Hasenkopf <ahasenkopf@suse.com>",
                    content="First enŧry",
                    timestamp=datetime(2017, 1, 1, 0, 0, 0, tzinfo=self.utc)
                ),
            ]

            self.assertEqual(len(cl.entries), 3)
            cl.write(path="/who/cares/test.changes")

            self.assertEqual(len(omock.mock_calls), 6)
            content = "".join(str(omock.mock_calls[x][1][0])
                              for x in range(2, 5))

            self.assertEqual(
                content,
                "-------------------------------------------------------------------\n"
                "Tue Jan 01 00:00:00 UTC 2019 - Andreas Hasenkopf <ahasenkopf@suse.com>\n\n"
                "Føø Bar\n\n"
                "-------------------------------------------------------------------\n"
                "Mon Jan 01 00:00:00 UTC 2018 - Andreas Pritschet <apritschet@suse.com>\n\n"
                "Hellø Wørld\n\n"
                "-------------------------------------------------------------------\n"
                "Sun Jan 01 00:00:00 UTC 2017 - Andreas Hasenkopf <ahasenkopf@suse.com>\n\n"
                "First enŧry\n\n"
            )

    def test_write_stringio(self):
        buffer = StringIO("")
        cl = ChangeLog()
        cl.entries = [Entry(
            packager="Andreas Hasenkopf <ahasenkopf@suse.com>",
            content="Føø Bar",
            timestamp=datetime(2019, 1, 1, 0, 0, 0, tzinfo=self.utc)
        )]

        self.assertEqual(len(cl.entries), 1)
        cl.write(path=buffer)
        buffer.seek(0)
        self.assertEqual(
                buffer.read(),
                "-------------------------------------------------------------------\n"
                "Tue Jan 01 00:00:00 UTC 2019 - Andreas Hasenkopf <ahasenkopf@suse.com>\n\n"
                "Føø Bar\n\n"
            )

    def test_write_append(self):
        with mock.patch("osctiny.utils.changelog.open",
                        mock.mock_open(read_data=SAMPLE_CHANGES_2),
                        create=True) as omock:
            self.prep_mock_open(omock)
            cl = ChangeLog.parse(path="/who/cares/test.changes",
                                 generative=False)

            cl.entries.append(Entry(
                packager="Andreas Hasenkopf <ahasenkopf@suse.com>",
                content="New entry at the top",
                timestamp=datetime(2020, 10, 31, 0, 0, 0, tzinfo=self.utc)
            ))
            cl.write(path="/who/cares/test.changes")

            write_calls = [x for x in omock.mock_calls if x[0] == "().write"]
            self.assertEqual(len(write_calls), 5)
            self.assertEqual(
                str(write_calls[0][1][0]),
                "-------------------------------------------------------------------\n"
                "Sat Oct 31 00:00:00 UTC 2020 - Andreas Hasenkopf <ahasenkopf@suse.com>\n\n"
                "New entry at the top\n\n"
            )

    def test_write_append_stringio(self):
        inbuff = StringIO(SAMPLE_CHANGES_2)
        outbuff = StringIO("")
        cl = ChangeLog.parse(path=inbuff, generative=False)

        cl.entries.append(Entry(
            packager="Andreas Hasenkopf <ahasenkopf@suse.com>",
            content="New entry at the top",
            timestamp=datetime(2020, 10, 31, 0, 0, 0, tzinfo=self.utc)
        ))
        cl.write(path=outbuff)

        outbuff.seek(0)
        self.assertEqual(
            len([x for x in outbuff.readlines() if x.startswith("-" * 67)]),
            5
        )

    def test_parse_missing_sep(self):
        buffer = StringIO(SAMPLE_CHANGES_3)
        cl = ChangeLog.parse(buffer, generative=False)

        self.assertIsInstance(cl.entries, list)
        self.assertEqual(len(cl.entries), 2)

    @mock.patch("warnings.warn")
    def test_parse_invalid_timestamp(self, wmock):
        buffer = StringIO(SAMPLE_CHANGES.replace("10:38:41", "10::41"))
        cl = ChangeLog.parse(buffer, generative=False)

        self.assertIsInstance(cl.entries, list)
        self.assertEqual(len(cl.entries), 2)
        self.assertIsInstance(cl.entries[0].timestamp, str)
        self.assertIsInstance(cl.entries[1].timestamp, datetime)

        self.assertEqual(wmock.call_count, 1)
        self.assertIn("Cannot parse changelog entry", wmock.call_args[0][0])


@mock.patch("osctiny.utils.conf._conf", new_callable=lambda: None, create=True)
@mock.patch("pathlib.Path.is_file", return_value=True)
class TestConfig(TestCase):
    def test_get_config_path(self, *_):
        with self.subTest("No env vars"):
            with mock.patch.dict(os.environ, values={}, clear=True):
                self.assertEqual(Path.home().joinpath(".oscrc"), get_config_path())

        with self.subTest("OSC_CONFIG"):
            osc_config = "/foo/bar/oscrc"
            with mock.patch.dict(os.environ, values={'OSC_CONFIG': osc_config}, clear=True):
                self.assertEqual(Path(osc_config), get_config_path())

    def test_get_credentials(self, *_):
        _, path1 = mkstemp()
        _, path2 = mkstemp()

        expected_insecure_credentials = ("my-dummy-user", "my-insecure-dummy-password", None)
        expected_secure_credentials = ('my-dummy-user', 'my-secure-dummy-password', None)

        with open(path1, "w") as handle:
            handle.write("[http://api.dummy-bs.org]\n")
            handle.write("user={}\npass={}\n".format(*expected_insecure_credentials))

        with open(path2, "w") as handle:
            handle.write("[general]\n")
            handle.write("apiurl=http://api.dummy-bs.org\n")
            handle.write("[http://api.dummy-bs.org]\n")
            handle.write("user={}\n".format(expected_secure_credentials[0]))
            handle.write("passx={}\n".format(b64encode(compress(
                expected_secure_credentials[1].encode("ascii")
            )).decode("ascii")))

        try:
            with self.subTest("No URL, no default"):
                with mock.patch.dict(os.environ, values={'OSC_CONFIG': path1}, clear=True):
                    self.assertRaises(ValueError, get_credentials)

            with self.subTest("Wrong URL"):
                with mock.patch.dict(os.environ, values={'OSC_CONFIG': path1}, clear=True):
                    self.assertRaises(ValueError, get_credentials, "http://google.de")

            with self.subTest("Wrong scheme"):
                with mock.patch.dict(os.environ, values={'OSC_CONFIG': path1}, clear=True):
                    self.assertRaises(ValueError, get_credentials, "https://api.dummy-bs.org")

            with self.subTest("Correct URL"):
                with mock.patch.dict(os.environ, values={'OSC_CONFIG': path1}, clear=True):
                    credentials = get_credentials("http://api.dummy-bs.org")
                self.assertEqual(expected_insecure_credentials, credentials)

            with self.subTest("No URL"):
                with mock.patch.dict(os.environ, values={'OSC_CONFIG': path2}, clear=True):
                    self.assertEqual(expected_secure_credentials, get_credentials())
        finally:
            os.remove(path1)
            os.remove(path2)



from unittest import skipUnless

from .test_search import TestSearch
from osctiny import Osc

try:
    import cachecontrol
except ImportError:
    with_cache = False
else:
    with_cache = True


@skipUnless(with_cache, "No cache module present, therefore not testing")
class TestSearch(TestSearch):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.osc = Osc(
            url="http://api.example.com",
            username="foobar",
            password="helloworld",
            cache=True
        )



# -*- coding: utf-8 -*-
import re
from urllib.parse import unquote_plus

import responses

from ..extensions import projects
from .base import OscTest, CallbackFactory


class BasicTest(OscTest):
    def test_handle_params(self):
        def _run(data, expected):
            handled = self.osc.handle_params(data)
            self.assertEqual(handled, expected)

        data = (
            (None, {}),
            (1, {}),
            ("hello world", b"hello world"),
            ("føø bær", b"f\xc3\xb8\xc3\xb8 b\xc3\xa6r"),
            (
                {"view": "xml", "withissues": 1},
                {b"view": b"xml", b"withissues": b"1"}
            ),
            (
                {"view": "xml", "withissues": True},
                {b"view": b"xml", b"withissues": b"1"}
            ),
            (
                {"view": "xml", "withissues": 0},
                {b"view": b"xml", b"withissues": b"0"}
            ),
            (
                {"view": "xml", "withissues": False},
                {b"view": b"xml", b"withissues": b"0"}
            ),
            (
                {"view": "xml", "withissues": None},
                {b"view": b"xml"}
            ),
        )

        for params, expected in data:
            with self.subTest(params):
                _run(params, expected)

    def test_attrib_regexp(self):
        def _run(attr, expected):
            match = projects.Project.attribute_pattern.match(attr)
            self.assertIsNotNone(match)
            self.assertEqual(match.groupdict(), expected)

        data = (
            ('foo', {'name': 'foo', 'prefix': None}),
            ('foo:bar', {'name': 'bar', 'prefix': 'foo'}),
            ('foo:bar:wørld', {'name': 'bar:wørld', 'prefix': 'foo'}),
        )

        for attr, expected in data:
            with self.subTest(attr):
                _run(attr, expected)

    @responses.activate
    def test_request_url_encode(self):
        pattern = re.compile(self.osc.url + r'file/(?P<filename>.*)')
        special_chars = ('#', '?')
        data = [
            ["Clean URL", "hello_world.txt"],
            ["URL with hashtag", "hełłø#wørłd.txt"],
            ["URL with question mark", "hello?wørłð.txt"],
            ["URL with ampersand", "hełłö&world.txt"]
        ]

        def callback(headers, params, request):
            match = pattern.match(request.url)
            self.assertIsNotNone(match)
            self.assertEqual(unquote_plus(match.group("filename")), filename)
            for special_c in special_chars:
                self.assertNotIn(special_c, match.group("filename"))
            return 200, headers, ""

        self.mock_request(
            method=responses.GET,
            url=pattern,
            callback=CallbackFactory(callback)
        )

        for name, filename in data:
            with self.subTest(name):
                self.osc.request(f"{self.osc.url}file/{filename}")



"""
Main API access
---------------
"""
from __future__ import unicode_literals

import typing
from io import BufferedReader, BytesIO, StringIO
import gc
import logging
from pathlib import Path
import re
from ssl import get_default_verify_paths
import time
from urllib.parse import quote
import warnings

# pylint: disable=no-name-in-module
from lxml.objectify import fromstring
from requests import Session, Request
from requests.auth import HTTPBasicAuth
from requests.cookies import RequestsCookieJar, cookiejar_from_dict
from requests.exceptions import ConnectionError as _ConnectionError

from .extensions.buildresults import Build
from .extensions.comments import Comment
from .extensions.distributions import Distribution
from .extensions.issues import Issue
from .extensions.origin import Origin
from .extensions.packages import Package
from .extensions.projects import Project
from .extensions.bs_requests import Request as BsRequest
from .extensions.search import Search
from .extensions.users import Group, Person
from .utils.auth import HttpSignatureAuth
from .utils.conf import get_credentials
from .utils.errors import OscError

try:
    from cachecontrol import CacheControl
except ImportError:
    CacheControl = None


# pylint: disable=too-many-instance-attributes,too-many-arguments
# pylint: disable=too-many-locals
class Osc:
    """
    Build service API client

    An instance of :py:class:`Osc` provides all the extensions and access to the
    API as attributes, e.g. to get a list of groups use:

    .. code-block:: python

        instance = Osc(*args, **kwargs)
        instance.groups.get_list()

    .. list-table:: Extensions
        :header-rows: 1

        * - Extension
          - Accessible through attribute
        * - :py:class:`osctiny.extensions.packages.Package`
          - :py:attr:`packages`
        * - :py:class:`osctiny.extensions.projects.Project`
          - :py:attr:`projects`
        * - :py:class:`osctiny.extensions.bs_requests.Request`
          - :py:attr:`requests`
        * - :py:class:`osctiny.extensions.search.Search`
          - :py:attr:`search`
        * - :py:class:`osctiny.extensions.users.Group`
          - :py:attr:`groups`
        * - :py:class:`osctiny.extensions.users.Person`
          - :py:attr:`users`
        * - :py:class:`osctiny.extensions.build.Build`
          - :py:attr:`build`
        * - :py:class:`osctiny.extensions.comments.Comment`
          - :py:attr:`comments`
        * - :py:class:`osctiny.extensions.distributions.Distribution`
          - :py:attr:`distributions`
        * - :py:class:`osctiny.extensions.origin.Origin`
          - :py:attr:`origins`

    :param url: API URL of a BuildService instance
    :param username: Username
    :param password: Password; this is either the user password or the SSH passphrase, if
                     ``ssh_key_file`` is defined
    :param verify: See `SSL Cert Verification`_ for more details
    :param cache: Store API responses in a cache
    :param ssh_key_file: Path to SSH private key file
    :raises osctiny.errors.OscError: if no credentials are provided

    .. versionadded:: 0.1.1
        The ``cache`` parameter and the ``build`` extension

    .. versionadded:: 0.1.8
        The ``comments`` extension

    .. versionadded:: 0.2.2
        The ``issues`` extension

    .. versionadded:: 0.2.3
        The ``distributions`` extension

    .. versionadded:: 0.3.0
        The ``origins`` extension

    .. versionchanged:: 0.4.0
        Raises an exception when no credentials are provided

    .. versionchanged:: 0.6.0
        Support for 2FA authentication (i.e. added the ``ssh_key_file`` parameter and changed the
        meaning of the ``password`` parameter

    .. _SSL Cert Verification:
        http://docs.python-requests.org/en/master/user/advanced/
        #ssl-cert-verification
    """
    url = 'https://api.opensuse.org'
    username = ''
    password = ''
    session = None
    _registered = {}
    default_timeout = (60, 300)
    default_connection_retries = 5
    default_retry_timeout = 5

    def __init__(self, url: str = None, username: typing.Optional[str] = None,
                 password: typing.Optional[str] = None, verify: typing.Optional[str] = None,
                 cache: bool = False,
                 ssh_key_file: typing.Optional[typing.Union[Path, str]] = None):
        # Basic URL and authentication settings
        self.url = url or self.url
        self.username = username or self.username
        self.password = password or self.password
        self.verify = verify
        self.cache = cache
        self.ssh_key = ssh_key_file
        if self.ssh_key is not None and not isinstance(self.ssh_key, Path):
            self.ssh_key = Path(self.ssh_key)

        if not self.username and not self.password and not self.ssh_key:
            try:
                self.username, self.password, self.ssh_key = get_credentials(self.url)
            except (ValueError, NotImplementedError, FileNotFoundError) as error:
                raise OscError from error

        # API endpoints
        self.build = Build(osc_obj=self)
        self.comments = Comment(osc_obj=self)
        self.distributions = Distribution(osc_obj=self)
        self.groups = Group(osc_obj=self)
        self.issues = Issue(osc_obj=self)
        self.origins = Origin(osc_obj=self)
        self.packages = Package(osc_obj=self)
        self.projects = Project(osc_obj=self)
        self.requests = BsRequest(osc_obj=self)
        self.search = Search(osc_obj=self)
        self.users = Person(osc_obj=self)

        self._session, self.session = None, None

    def __del__(self):
        # Just in case ;-)
        gc.collect()

    @property
    def cookies(self) -> RequestsCookieJar:
        """
        Access session cookies
        """
        if self._session is None:
            self._init_session()

        return self.session.cookies

    @cookies.setter
    def cookies(self, value: RequestsCookieJar):
        if not isinstance(value, (RequestsCookieJar, dict)):
            raise TypeError(f"Expected a cookie jar or dict. Got instead: {type(value)}")

        if self._session is None:
            self._init_session()

        if isinstance(value, RequestsCookieJar):
            self._session.cookies = value
        else:
            self._session.cookies = cookiejar_from_dict(value)

    def _init_session(self):
        """
        Lazy session initialization
        """
        self._session = Session()
        self._session.verify = self.verify or get_default_verify_paths().capath

        if self.ssh_key is not None:
            self._session.auth = HttpSignatureAuth(username=self.username, password=self.password,
                                                   ssh_key_file=self.ssh_key)
        else:
            self._session.auth = HTTPBasicAuth(self.username, self.password)

        # Cache
        if self.cache:
            # pylint: disable=broad-except
            try:
                self.session = CacheControl(self._session)
            except Exception as error:
                self.session = self._session
                warnings.warn("Cannot use the cache: {}".format(error),
                              RuntimeWarning)
        else:
            self.session = self._session

    def request(self, url, method="GET", stream=False, data=None, params=None,
                raise_for_status=True, timeout=None):
        """
        Perform HTTP(S) request

        ``data`` is URL-encoded and passed on as GET parameters. If ``data`` is
        a dictionary and contains a key ``comment``, this value is passed on as
        a POST parameter.

        If ``stream`` is True, the server response does not get cached because
        the returned file might be large or huge.

        if ``raise_for_status`` is True, the used ``requests`` framework will
        raise an exception for occured errors.

        .. versionchanged:: 0.1.2
            Added parameters `raise_for_status`

        .. versionchanged:: 0.1.3

            * Added parameter `timeout`
            * Transfer data as GET parameters (except for comments and texts)

        .. versionchanged:: 0.1.5
            Retry sending the request, if the remote host disconnects

        .. versionadded:: 0.1.7
            Added parameter `params`

        .. versionchanged:: 0.5.0
            Added logging of request/response

        :param url: Full URL
        :param method: HTTP method
        :param stream: Delayed access, see `Body Content Workflow`_
        :param data: Data to be included as POST parameters in request
        :param params: Additional GET parameters to be included in request
        :param raise_for_status: See `requests.Response.raise_for_status`_
        :param timeout: Request timeout. See `Timeouts`_
        :return: :py:class:`requests.Response`

        .. _Body Content Workflow:
            http://docs.python-requests.org/en/master/user/advanced/
            #body-content-workflow

        .. _requests.Response.raise_for_status:
            https://2.python-requests.org/en/master/api/
            #requests.Response.raise_for_status

        .. _Timeouts:
            https://2.python-requests.org/en/master/user/advanced/#timeouts
        """
        timeout = timeout or self.default_timeout
        if self._session is None:
            self._init_session()

        if stream:
            session = self._session
        else:
            session = self.session

        req = Request(
            method,
            url.replace("#", quote("#")).replace("?", quote("?")),
            data=self.handle_params(data),
            params=self.handle_params(params)
        )
        prepped_req = session.prepare_request(req)
        prepped_req.headers['Content-Type'] = "application/octet-stream"
        prepped_req.headers['Accept'] = "application/xml"
        settings = session.merge_environment_settings(
            prepped_req.url, {}, None, None, None
        )
        settings["stream"] = stream
        if timeout:
            settings["timeout"] = timeout

        logger = logging.getLogger("osctiny.request")

        for i in range(self.default_connection_retries, -1, -1):
            logger.info("Requested URL: %s", prepped_req.url)
            logger.debug("Sent data:\n%s\n---",
                         "\n".join(f"{k}: {v}" for k, v in req.data.items())
                         if isinstance(req.data, dict) else req.data)
            logger.debug("Sent parameters:\n%s\n---",
                         "\n".join(f"{k}: {v}" for k, v in req.params.items()))
            try:
                response = session.send(prepped_req, **settings)
            except _ConnectionError as error:
                warnings.warn("Problem connecting to server: {}".format(error))
                log_method = logger.error if i < 1 else logger.warning
                log_method("Request failed: %s", error)
                if i < 1:
                    raise
                logger.debug("Retrying request in %d seconds", self.default_retry_timeout)
                time.sleep(self.default_retry_timeout)
            else:
                logger.info("Server replied with status %d", response.status_code)
                logger.debug("Response headers:\n%s\n---",
                             "\n".join(f"{k}: {v}" for k, v in response.headers.items()))
                logger.debug("Response content:\n%s\n---", response.text)
                if raise_for_status:
                    response.raise_for_status()
                return response

        return None

    @staticmethod
    def handle_params(params):
        """
        Translate request parameters to API conform format

        .. note:: The build service does not handle URL-encoded Unicode well.
                  Therefore parameters are encoded as ``bytes``.

        :param params: Request parameter
        :type params: dict or str or io.BufferedReader
        :return: converted data ready to be used in HTTP request
        :rtype: dict or bytes
        """
        if isinstance(params, bytes):
            return params

        if isinstance(params, str):
            return params.encode('utf-8')

        if isinstance(params, StringIO):
            params.seek(0)
            return params.read().encode('utf-8')

        if isinstance(params, (BufferedReader, BytesIO)):
            params.seek(0)
            return params

        if not isinstance(params, dict):
            return {}

        for key in params:
            if isinstance(params[key], bool):
                params[key] = '1' if params[key] else '0'

        return {key.encode(): str(value).encode()
                for key, value in params.items()
                if value is not None}

    def get_objectified_xml(self, response):
        """
        Return API response as an XML object

        .. versionchanged:: 0.1.6

            Allow parsing of "huge" XML inputs

        .. versionchanged:: 0.2.4

            Allow ``response`` to be a string

        :param response: An API response or XML string
        :rtype response: :py:class:`requests.Response`
        :return: :py:class:`lxml.objectify.ObjectifiedElement`
        """
        if isinstance(response, str):
            text = response
        else:
            text = response.text

        try:
            return fromstring(text)
        except ValueError:
            # Just in case OBS returns a Unicode string with encoding
            # declaration
            if isinstance(text, str) and \
                    "encoding=" in text:
                return fromstring(
                    re.sub(r'encoding="[^"]+"', "", text)
                )

            # This might be something else
            raise



"""
Backports of handy new features
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. versionadded:: 0.3.0
"""
try:
    # pylint: disable=unused-import
    from functools import lru_cache
except ImportError:
    # Whoever had the grandiose idea to backport this to Python2?
    # pylint: disable=unused-argument
    def lru_cache(*args, **kwargs):
        """Dummy wrapper"""
        def wrapper(fun):
            return fun

        return wrapper



# -*- coding: utf-8 -*-
from io import StringIO, BytesIO, IOBase
import re
from unittest import skip

import responses

from .base import OscTest, CallbackFactory
from ..utils.errors import OscError


class TestPackage(OscTest):
    @responses.activate
    def test_get_files(self):
        def callback(headers, params, request):
            status = 200
            body = """
                <directory name="python.8549" rev="1" vrev="28.17" 
                    srcmd5="9498d242f91b858372705d1bb4e26e1a">
                  <entry name="CVE-2017-18207.patch" 
                    md5="d1e5e39cfdf5087933cabf7d7b541158" size="900" 
                    mtime="1528391213" />
                  <entry name="Python-2.7.13.tar.xz" 
                    md5="53b43534153bb2a0363f08bae8b9d990" size="12495628"
                    mtime="1484132166" />
                  <entry name="Python-2.7.13.tar.xz.asc" 
                    md5="115f2de1793fa46a382f2a9f4e11b285" size="801" 
                    mtime="1484132166" />
                  <entry name="README.SUSE" 
                    md5="4a5a6c13a5b163d2e763b0d45f64c051" size="735" 
                    mtime="1216853740" />
                  <entry name="python-2.7.13-docs-pdf-a4.tar.bz2" 
                    md5="ccc87ad010f28d926ca09b1d5a55f4b0" size="10712181" 
                    mtime="1484132167" />
                  <entry name="python-base.spec" 
                    md5="af86457b494a667420e971e2e0febfcf" size="18454" 
                    mtime="1537948380" />
                  <entry name="python-doc.changes" 
                    md5="d5b75815a9455e231cc12d0aac045f9c" size="7290" 
                    mtime="1537948380" />
                  <entry name="python-doc.spec" 
                    md5="832b21401d9951d3b6e717e80944bd2b" size="6522" 
                    mtime="1537948381" />
                  <entry name="python.changes" 
                    md5="d479c6cd83559b1eed7d909087cf4785" size="54144" 
                    mtime="1537948381" />
                  <entry name="python.spec" 
                    md5="812c70b56ffd92d060bcf8b02ba479ff" size="20189" 
                    mtime="1537948382" />
                </directory>
            """
            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=self.osc.url + '/source/SUSE:SLE-12-SP1:Update/python.8549',
            callback=CallbackFactory(callback)
        )

        response = self.osc.packages.get_files(
            "SUSE:SLE-12-SP1:Update", "python.8549"
        )
        self.assertEqual(response.tag, "directory")
        self.assertEqual(response.countchildren(), 10)

    @responses.activate
    def test_get_list(self):
        def callback(headers, params, request):
            status = 200
            body = """
                <directory count="14">
                  <entry name="SAPHanaSR"/>
                  <entry name="SAPHanaSR.4926"/>
                  <entry name="SAPHanaSR.7820"/>
                  <entry name="SUSEConnect"/>
                  <entry name="SUSEConnect.1732"/>
                  <entry name="SUSEConnect.1892"/>
                  <entry name="SUSEConnect.2196"/>
                  <entry name="SUSEConnect.2374"/>
                  <entry name="SUSEConnect.4293"/>
                  <entry name="SUSEConnect.4515"/>
                  <entry name="SUSEConnect.4773"/>
                  <entry name="SUSEConnect.7260"/>
                  <entry name="SUSEConnect.8868"/>
                  <entry name="SUSEConnect.9195"/>
                </directory>
            """
            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=self.osc.url + '/source/SUSE:SLE-12-SP1:Update/python.8549',
            callback=CallbackFactory(callback)
        )

        response = self.osc.packages.get_files(
            "SUSE:SLE-12-SP1:Update", "python.8549"
        )
        self.assertEqual(response.tag, "directory")
        self.assertEqual(response.countchildren(), 14)

    @responses.activate
    def test_get_meta(self):
        def callback(headers, params, request):
            status = 200
            body = """
            <package name="python.8549" project="SUSE:SLE-12-SP1:Update">
              <title>Python Interpreter</title>
              <description>Python is an interpreted, object-oriented programming
              language, and is often compared to Tcl, Perl, Scheme, or Java. You
              can find an overview of Python in the documentation and tutorials
              included in the python-doc (HTML) or python-doc-pdf (PDF) 
              packages.
            
              If you want to install third party modules using distutils, you 
              need to install python-devel package.</description>
              <releasename>python</releasename>
            </package>
            """
            if params.get("view", None) == "blame":
                body = """
   1 (foo__bar     2018-10-29 16:28:55     1) <package name="python.8549" project="SUSE:SLE-12-SP1:Update">
   1 (foo__bar     2018-10-29 16:28:55     2)   <title>Python Interpreter</title>
   1 (foo__bar     2018-10-29 16:28:55     3)   <description>Python is an interpreted, object-oriented programming language, and is
   1 (foo__bar     2018-10-29 16:28:55     4) often compared to Tcl, Perl, Scheme, or Java.  You can find an overview
   1 (foo__bar     2018-10-29 16:28:55     5) of Python in the documentation and tutorials included in the python-doc
   1 (foo__bar     2018-10-29 16:28:55     6) (HTML) or python-doc-pdf (PDF) packages.
   1 (foo__bar     2018-10-29 16:28:55     7) 
   1 (foo__bar     2018-10-29 16:28:55     8) If you want to install third party modules using distutils, you need to
   1 (foo__bar     2018-10-29 16:28:55     9) install python-devel package.</description>
   1 (foo__bar     2018-10-29 16:28:55    10)   <releasename>python</releasename>
   1 (foo__bar     2018-10-29 16:28:55    11) </package>
                """

            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=self.osc.url + '/source/SUSE:SLE-12-SP1:Update/python.8549/'
                               '_meta',
            callback=CallbackFactory(callback)
        )

        with self.subTest("without blame"):
            response = self.osc.packages.get_meta(
                "SUSE:SLE-12-SP1:Update", "python.8549"
            )
            self.assertEqual(response.tag, "package")
            self.assertEqual(
                response.xpath("./title")[0].text,
                "Python Interpreter"
            )

        with self.subTest("with blame"):
            response = self.osc.packages.get_meta(
                "SUSE:SLE-12-SP1:Update", "python.8549", blame=True
            )
            self.assertTrue(isinstance(response, str))

    @skip("No test data available")
    @responses.activate
    def test_get_attribute(self):
        def callback(headers, params, request):
            status = 200
            body = """</attributes>"""
            headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
            return status, headers, body

        self.mock_request(
            method=responses.GET,
            url=self.osc.url + '/source/SUSE:SLE-12-SP1:Update/python.8549/'
                               '_attribute',
            callback=CallbackFactory(callback)
        )

    @responses.activate
    def test_get_history(self):
        self.mock_request(
            method=responses.GET,
            url=self.osc.url + '/source/SUSE:SLE-12-SP1:Update/python.8549/'
                               '_history',
            body="""
                <revisionlist>
                  <revision rev="1" vrev="1">
                    <srcmd5>b9b258599bb67a2a3d396b1515cabeab</srcmd5>
                    <version>unknown</version>
                    <time>1514367595</time>
                    <user>Fȱȱ Bar</user>
                    <comment>
                      Set link tȱ grub2.5745 via maintenance_release request
                    </comment>
                    <requestid>148865</requestid>
                  </revision>
                  <revision rev="2" vrev="2">
                    <srcmd5>9f5e43584f67e2a301b71b63bdf8e2e1</srcmd5>
                    <version>unknown</version>
                    <time>1520336862</time>
                    <user>HȨllȱ Wȱrld</user>
                    <comment>
                      Set link tȱ grub2.6584 via maintenance_release request
                    </comment>
                    <requestid>154349</requestid>
                  </revision>
                </revisionlist>
            """
        )

        response = self.osc.packages.get_history(
            "SUSE:SLE-12-SP1:Update", "python.8549"
        )
        self.assertEqual(response.tag, "revisionlist")
        self.assertEqual(
            len(response.xpath("./revision")), 2
        )

    @responses.activate
    def test_cmd(self):
        self.mock_request(
            method=responses.POST,
            url=self.osc.url + '/source/SUSE:SLE-12-SP1:Update/python.8549',
            body="""
                +==== //tools/python/2.6.2/src/base/Modules/_ctypes/libffi/src/sparc/ffi.c#1 - /home/build/clifford/gpdb/tools/python/2.6.2/src/base/Modules/_ctypes/libffi/src/sparc/ffi.c ====
                +---
                + Modules/_ctypes/libffi/src/sparc/ffi.c |    5 +++++
                + 1 file changed, 5 insertions(+)
                +
                +--- a/Modules/_ctypes/libffi/src/sparc/ffi.c
                ++++ b/Modules/_ctypes/libffi/src/sparc/ffi.c
                +@@ -652,6 +652,11 @@
                + 	}
                +       else
                + 	{
                ++#if FFI_TYPE_LONGDOUBLE != FFI_TYPE_DOUBLE
                ++         /* SparcV9 long double is 16-byte aligned; skip arg if necessary */
                ++         if (arg_types[i]->type == FFI_TYPE_LONGDOUBLE && (argn & 1))
                ++           argn++;
                ++#endif
                + 	  /* Right-justify.  */
                + 	  argn += ALIGN(arg_types[i]->size, FFI_SIZEOF_ARG) / FFI_SIZEOF_ARG;
                + 
            """
        )

        response = self.osc.packages.cmd(
            "SUSE:SLE-12-SP1:Update", "python.8549", "diff"
        )
        self.assertTrue(isinstance(response, str))
        self.assertIn(
            "++#if FFI_TYPE_LONGDOUBLE != FFI_TYPE_DOUBLE", response
        )

    @responses.activate
    def test_set_meta(self):
        bodies = []

        def callback(headers, params, request):
            bodies.append(request.body)
            status, body = 200, ""

            return status, headers, body

        self.mock_request(
            method=responses.PUT,
            url=re.compile(self.osc.url + '/source/(?P<project>[^/]+)/'
                                          '(?P<package>[^/]+)/_meta'),
            callback=CallbackFactory(callback)
        )

        data = (
            (
                {
                    'project': "test:project",
                    'package': "test.package",
                    'title': "test",
                    'description': "foo"
                },
                b'<package><title>test</title><description>foo</description>'
                b'</package>'
            ),
            (
                {
                    'project': "test:project",
                    'package': "test.package",
                },
                b'<package><title/><description/></package>'
            ),
            (
                {
                    'project': "test:project",
                    'package': "test.package",
                    'meta': """
                    <package name="test.package" project="test:project">
                      <title/>
                      <description/>
                      <build>
                        <enable repository="openSUSE_Leap_15.0"/>
                        <disable arch="i586"/>
                      </build>
                    </package>
                    """
                },
                b'<package name="test.package" project="test:project"><title/>'
                b'<description/><build>'
                b'<enable repository="openSUSE_Leap_15.0"/>'
                b'<disable arch="i586"/></build></package>'
            ),
            (
                {
                    'project': "test:project",
                    'package': "test.package",
                    'title': 'foo',
                    'description': 'bar',
                    'meta': """
                    <package name="test.package" project="test:project">
                      <title/>
                      <description/>
                      <build>
                        <enable repository="openSUSE_Leap_15.0"/>
                        <disable arch="i586"/>
                      </build>
                    </package>
                    """
                },
                b'<package name="test.package" project="test:project">'
                b'<title>foo</title><description>bar</description><build>'
                b'<enable repository="openSUSE_Leap_15.0"/>'
                b'<disable arch="i586"/></build></package>'
            ),
        )

        for params, expected in data:
            with self.subTest():
                self.osc.packages.set_meta(**params)
                self.assertEqual(bodies[-1], expected)

    @responses.activate
    def test_push_file(self):
        content = """
        ლ(ಠ益ಠ)ლ            ლ(ಠ益ಠ)ლ
        Lorem ipsum dolor sit amet,
        consectetur adipiscing elit.
        Vestibulum id enim 
        fermentum, lobortis urna
        quis, convallis justo. 
        ლ(ಠ益ಠ)ლ            ლ(ಠ益ಠ)ლ 
        """
        bodies = []
        received_params = []

        def callback(headers, params, request):
            if isinstance(request.body, IOBase):
                request.body.seek(0)
                bodies.append(request.body.read())
            else:
                bodies.append(request.body)
            status, body = 200, ""
            received_params.append(params)

            return status, headers, body

        self.mock_request(
            method=responses.PUT,
            url=re.compile(
                self.osc.url + '/source/(?P<project>[^/]+)/'
                               '(?P<package>[^/]+)/(?P<filename>.+)'
            ),
            callback=CallbackFactory(callback)
        )

        with self.subTest("as unicode"):
            self.osc.packages.push_file("prj", "pkg", "readme.txt", content)
            self.assertEqual(bodies[-1], content.encode('utf-8'))

        with self.subTest("as bytes"):
            self.osc.packages.push_file("prj", "pkg", "readme.txt",
                                        content.encode('utf-8'))
            self.assertEqual(bodies[-1], content.encode('utf-8'))

        with self.subTest("as StringIO"):
            self.osc.packages.push_file("prj", "pkg", "readme.txt",
                                        StringIO(content))
            self.assertEqual(bodies[-1], content.encode('utf-8'))

        with self.subTest("as BytesIO"):
            self.osc.packages.push_file("prj", "pkg", "readme.txt",
                                        BytesIO(content.encode('utf-8')))
            self.assertEqual(bodies[-1], content.encode('utf-8'))

        with self.subTest("with comment"):
            the_comment = "This is a comment"
            self.osc.packages.push_file("prj", "pkg", "readme.txt", content,
                                        comment=the_comment)
            self.assertEqual(received_params[-1]["comment"], [the_comment])


    @responses.activate
    def test_aggregate(self):
        put_called = []

        def exists_callback(headers, params, request):
            status, body = 404, ""
            if "exists" in request.url:
                if "_aggregate" in request.url and "already.agged" not in request.url:
                    status = 404
                else:
                    status = 200
            return status, headers, body

        def put_callback(headers, params, request):
            put_called.append({'request': request, 'params': params})
            status, body = 200, ""
            return status, headers, body

        def meta_callback(headers, params, request):
            status = 200
            body = """<package><title/><description/></package>"""
            return status, headers, body

        self.mock_request(
            method=responses.HEAD,
            url=re.compile(
                self.osc.url + '/source/.*'
            ),
            callback=CallbackFactory(exists_callback)
        )
        self.mock_request(
            method=responses.PUT,
            url=re.compile(
                self.osc.url + '/source/.*'
            ),
            callback=CallbackFactory(put_callback)
        )
        self.mock_request(
            method=responses.GET,
            url=re.compile(
                self.osc.url + '/source/.+/_meta'
            ),
            callback=CallbackFactory(meta_callback)
        )

        with self.subTest("identical package"):
            self.assertRaises(
                OscError,
                self.osc.packages.aggregate,
                "test:project", "test.package",
                "test:project", "test.package",
            )

        with self.subTest("non-existing package"):
            self.assertRaises(
                OscError,
                self.osc.packages.aggregate,
                "test:project", "test.package",
                "test:project:2", "test.package",
            )

        with self.subTest("already existing aggregate"):
            self.assertRaises(
                OscError,
                self.osc.packages.aggregate,
                "test:project:exists", "test.package",
                "test:project2:exists", "already.agged",
            )

        with self.subTest("non-existing target package"):
            old_len = len(put_called)
            self.osc.packages.aggregate(
                "test:project:exists", "test.package",
                "test:project2:foo", "test.pkg",
            )
            self.assertEqual(len(put_called), old_len + 2)

        with self.subTest("existing target package"):
            old_len = len(put_called)
            self.osc.packages.aggregate(
                "test:project:exists", "test.package",
                "test:project2:exists", "test.pkg",
            )
            self.assertEqual(len(put_called), old_len + 1)



"""
Requests extension
------------------
"""
from __future__ import unicode_literals
from six.moves.urllib.parse import urljoin
from six import text_type

from ..utils.base import ExtensionBase


class Request(ExtensionBase):
    """
    The BuildService request API is accessible through this object.
    """
    base_path = "/request/"

    @staticmethod
    def _validate_id(request_id):
        request_id = text_type(request_id)
        if not request_id.isnumeric():
            raise ValueError(
                "Request ID must be numeric! Got instead: {}".format(request_id)
            )
        return request_id

    def get_list(self, **params):
        """
        Get a list or request objects

        :param params: see https://build.opensuse.org/apidocs/index#73
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path),
            method="GET",
            params=params
        )

        return self.osc.get_objectified_xml(response)

    def get(self, request_id, withhistory=False, withfullhistory=False):
        """
        Get one request object

        :param request_id: ID of the request
        :param withhistory: includes the request history in result
        :type withhistory: bool
        :param withfullhistory: includes the request and review history in
                                result
        :type withfullhistory: bool
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        request_id = self._validate_id(request_id)
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path + request_id),
            method="GET",
            params={
                'withhistory': withhistory,
                'withfullhistory': withfullhistory
            }
        )

        return self.osc.get_objectified_xml(response)

    def cmd(self, request_id, cmd="diff", **kwargs):
        """
        Get the result of the specified command

        Available commands:

        * `diff`: Shows the diff of all affected packages.
        * `addreview`: Assign a user or group to review
        * `changereviewstate`: Accept/Decline a review

        :param request_id: ID of the request
        :param cmd: Name of the command
        :param view: One of: ``xml`` or nothing
        :param unified: ???
        :param missingok: ???
        :param filelimit: ???
        :param expand: ???
        :param withissues: ???
        :return: plain text
        :rtype: str

        .. versionchanged:: 0.1.3

            * Added ``addreview`` to list of allowed commands
            * Added validation for arguments of command ``changereviewstate``
        """
        allowed_cmds = ['diff', 'changereviewstate', 'addreview']
        allowed_review_states = ['new', 'accepted', 'declined', 'deleted',
                                 'revoked', 'superseded']
        if cmd not in allowed_cmds:
            raise ValueError("Invalid command: '{}'. Use one of: {}".format(
                cmd, ", ".join(allowed_cmds)
            ))

        if cmd == "changereviewstate"\
                and kwargs.get("newstate", None) not in allowed_review_states:
            raise ValueError(
                "Invalid review state: '{}'. Use one of: {}".format(
                    kwargs.get("newstate", None), allowed_review_states
                )
            )

        kwargs["cmd"] = cmd
        request_id = self._validate_id(request_id)
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path + request_id),
            method="POST",
            params=kwargs
        )

        if kwargs.get("view", "plain") == "xml":
            return self.osc.get_objectified_xml(response)
        return response.text

    def add_comment(self, request_id, comment, parent_id=None):
        """
        Add a comment to a request

        .. versionadded:: 0.1.1

        .. versionchanged:: 0.1.8
            Use internally :py:class:`osctiny.comments.Comment.add`

        :param request_id: ID of the request
        :param comment: Comment to be added
        :param parent_id: ID of parent comment. Default: ``None``
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        return self.osc.comments.add(
            obj_type="request",
            ids=(request_id,),
            comment=comment,
            parent_id=parent_id
        )

    def get_comments(self, request_id):
        """
        Get a list of comments for request

        .. versionchanged:: 0.1.8
            Use internally :py:class:`osctiny.comments.Comment.get`

        :param request_id: ID of the request
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        return self.osc.comments.get(
            obj_type="request",
            ids=(request_id,)
        )



from io import IOBase
from unittest import TestCase
from urllib.parse import parse_qs, urlparse

import responses

from osctiny import Osc


class CallbackFactory:
    """
    Wrapper for callback functions

    Callback functions are expected to accept three parameters and return
    three values:

    :param headers: Predefined headers
    :type headers: dict
    :param params: Parsed querystring
    :param request: Actual request object
    :type: requests.Request
    :type params: dict
    :return: (HTTP status, headers, JSON body)
    :rtype: (int, dict, str)
    """
    def __init__(self, callback):
        if not callable(callback):
            raise TypeError("Callback needs to be callable")
        self.callback = callback

    def __call__(self, request):
        body = request.body
        if isinstance(body, IOBase):
            body.seek(0)
            body = body.read()
        if hasattr(body, "decode"):
            body = body.decode('utf-8')
        params = parse_qs(body) if body else {}
        parsed = urlparse(request.url)
        params.update(parse_qs(parsed.query))
        headers = {
            "Cache-Control": "max-age=0, private, must-revalidate",
            "Connection": "Keep-Alive",
            "Referrer-Policy": "strict-origin-when-cross-origin",
            "Server": "Apache/2.4.33 (Linux/SUSE)",
            "X-Content-Type-Options": "nosniff",
            "X-Download-Options": "noopen",
            "X-Frame-Options": "SAMEORIGIN",
            "X-Request-Id": "6ac6301b-0460-486f-b273-7308ddf673b1",
            "X-XSS-Protection": "1; mode=block"
        }
        return self.callback(headers, params, request)


class OscTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super(OscTest, cls).setUpClass()
        cls.osc = Osc(
            url="http://api.example.com",
            username="foobar",
            password="helloworld",
        )
        cls.osc.default_connection_retries = 0
        cls.osc.default_retry_timeout = 0

    @staticmethod
    def mock_request(**kwargs):
        kwargs.setdefault("content_type", "application/xml")

        if "callback" in kwargs:
            responses.add_callback(**kwargs)
        else:
            responses.add(**kwargs)



# -*- coding: utf-8 -*-
import re
from urllib.parse import urlparse, parse_qs

from lxml.objectify import ObjectifiedElement
import responses

from .base import OscTest, CallbackFactory


def callback(headers, params, request):
    status = 500
    body = ""
    parsed = urlparse(request.url)
    params.update(parse_qs(parsed.query))

    if re.search("comments/request/30902", request.url):
        status = 200
        if request.method == "GET":
            body = """
            <comments request="30902">
              <comment who="apritschet" when="2019-04-18 15:31:19 UTC" 
                       id="1571414">test hello world 2</comment>
              <comment who="apritschet" when="2019-04-18 15:32:42 UTC" 
                       id="1571422">foo bar</comment>
            </comments>"""
        elif request.method == "POST":
            body = """
            <status code="ok" class="mozwebext">
              <summary>Ok</summary>
              <details>Operation successfull.</details>
            </status>
            """
    elif request.method == "GET" and re.search("/request/30902/?", request.url):
        status = 200
        body = """
            <request id="30902" creator="nemo">
          <action type="submit">
            <source project="SUSE:Factory:Head" 
              package="perl-XML-DOM-XPath" rev="3"/>
            <target project="SUSE:SLE-12:GA" 
              package="perl-XML-DOM-XPath"/>
            <acceptinfo rev="1" 
              srcmd5="b3b89e4efa25d04f814bcce615455ac7" 
              osrcmd5="d41d8cd98f00b204e9800998ecf8427e"/>
          </action>
          <state name="accepted" who="foobar" 
            when="2014-01-24T13:11:52">
            <comment>Accepted submit request 30902 from user helloworld
            </comment>
          </state>
          <review state="accepted" when="2014-01-22T17:51:08" 
              who="foobar" by_package="perl-XML-DOM-XPath" 
              by_project="SUSE:Factory:Head">
            <comment>okay</comment>
            <history who="foobar" when="2014-01-23T12:30:39">
              <description>Review got accepted</description>
              <comment>okay</comment>
            </history>
          </review>
          <review state="accepted" when="2014-01-22T17:51:08" 
              who="helloworld" by_group="maintenance-team">
            <comment>Accept manually without review.</comment>
            <history who="helloworld" when="2014-01-22T18:12:29">
              <description>Review got accepted</description>
              <comment>Accept manually without review.</comment>
            </history>
          </review>
          <review state="accepted" when="2014-01-22T17:55:07" 
              who="echo" by_group="legal-team">
            <comment/>
            <history who="echo" when="2014-01-23T19:24:37">
              <description>Review got accepted</description>
            </history>
          </review>
          <description>Required package for FATE#315181 (virt-v2v)
          </description>
        """
        if params.get("withhistory", ['0']) == ['1'] \
                or params.get("withfullhistory", ['0']) == ['1']:
            body += """
                <history who="foo" when="2014-01-22T17:51:08">
                  <description>Request created</description>
                  <comment>Required package for FATE#315181 (virt-v2v)</comment>
                </history>
                <history who="bar" when="2014-01-22T17:55:07">
                  <description>Request got a new review request</description>
                  <comment>{"delegate": "Logic only implemented for SUSE"}
                  </comment>
                </history>
                <history who="bar" when="2014-01-23T19:24:37">
                  <description>Request got reviewed</description>
                </history>
                <history who="hello" when="2014-01-24T13:11:52">
                  <description>Request got accepted</description>
                  <comment>Accepted submit request 30902 from user nemo
                  </comment>
                </history>
            """
        if params.get("withfullhistory", ['0']) == ['1']:
            body += """
                  <history who="acceptor" when="2014-01-23T19:24:37">
                    <description>Review got accepted</description>
                  </history>
            """
        body += "</request>"
    elif re.search("request/?$", request.url):
        status = 200
        body = """
            <directory>
              <entry name="179682"/>
              <entry name="179683"/>
              <entry name="179684"/>
              <entry name="179685"/>
              <entry name="179686"/>
              <entry name="179687"/>
              <entry name="179688"/>
              <entry name="179689"/>
              <entry name="179690"/>
              <entry name="179691"/>
              <entry name="179692"/>
              <entry name="179693"/>
            </directory>
        """
    elif (request.method == "GET" and re.search(
            "request/30902/?cmd=diff$", request.url
    )) or (request.method == "POST" and "diff" in params.get("cmd", [])):
        status = 200
        if "xml" in params.get("view", ["plain"]):
            body = """<request id="30902" actions="0">
  <action type="submit">
  <source project="SUSE:Factory:Head" package="perl-XML-DOM-XPath" rev="3"/>
  <target project="SUSE:SLE-12:GA" package="perl-XML-DOM-XPath"/>
  <acceptinfo rev="1" srcmd5="b3b89e4efa25d04f814bcce615455ac7" osrcmd5="d41d8cd98f00b204e9800998ecf8427e"/>
<sourcediff key="0de9f643081bd0b6e287be4e72c4f388">
  <old project="SUSE:SLE-12:GA" package="perl-XML-DOM-XPath" rev="d41d8cd98f00b204e9800998ecf8427e" srcmd5="d41d8cd98f00b204e9800998ecf8427e"/>
  <new project="SUSE:SLE-12:GA" package="perl-XML-DOM-XPath" rev="b3b89e4efa25d04f814bcce615455ac7" srcmd5="b3b89e4efa25d04f814bcce615455ac7"/>
  <files>
    <file state="added">
      <new name="XML-DOM-XPath-0.14.tar.gz" md5="51a40df96c2f92829e1a4f84782fa75e" size="12410"/>
      <diff binary="1" lines="0"/>
    </file>
    <file state="added">
      <new name="perl-XML-DOM-XPath.changes" md5="350164d3d923003682a645e8b1790edb" size="721"/>
      <diff lines="23">@@ -0,0 +1,22 @@
+-------------------------------------------------------------------
+Tue Aug  6 22:46:18 UTC 2013 - rcurtis@suse.com
+
+- Updated spec file to replace perl_requires macro with preferred SUSE Requires:  perl &gt;= x.x format
+
+-------------------------------------------------------------------
+Mon Aug  5 10:10:37 UTC 2013 - cfarrell@suse.com
+
+- license update: Artistic-1.0 or GPL-1.0+
+  SPDX syntax
+
+-------------------------------------------------------------------
+Wed Dec  1 13:36:09 UTC 2010 - coolo@novell.com
+
+- switch to perl_requires macro
+
+-------------------------------------------------------------------
+Wed Aug  4 12:51:34 UTC 2010 - chris@computersalat.de
+
+- initial package 0.14
+  * created by cpanspec 1.78
+
</diff>
    </file>
    <file state="added">
      <new name="perl-XML-DOM-XPath.spec" md5="584694dc246c40f449d20b9e158fda20" size="2300"/>
      <diff lines="79">@@ -0,0 +1,78 @@
+#
+# spec file for package perl-XML-DOM-XPath
+#
+# Copyright (c) 2013 SUSE LINUX Products GmbH, Nuernberg, Germany.
+#
+# All modifications and additions to the file contributed by third parties
+# remain the property of their copyright owners, unless otherwise agreed
+# upon. The license for this file, and modifications and additions to the
+# file, is the same license as for the pristine package itself (unless the
+# license for the pristine package is not an Open Source License, in which
+# case the license is the MIT License). An "Open Source License" is a
+# license that conforms to the Open Source Definition (Version 1.9)
+# published by the Open Source Initiative.
+
+# Please submit bugfixes or comments via http://bugs.opensuse.org/
+#
+
+
+%bcond_with pod
+
+Name:           perl-XML-DOM-XPath
+%define cpan_name XML-DOM-XPath
+Summary:        Perl extension to add XPath support to XML::DOM, using XML::XPath engine
+License:        Artistic-1.0 or GPL-1.0+
+Group:          Development/Libraries/Perl
+Version:        0.14
+Release:        0
+Url:            http://search.cpan.org/dist/XML-DOM-XPath/
+Source:         http://www.cpan.org/modules/by-module/XML/XML-DOM-XPath-%{version}.tar.gz
+BuildArch:      noarch
+BuildRoot:      %{_tmppath}/%{name}-%{version}-build
+Requires:       perl &gt;= 5.16
+BuildRequires:  perl &gt;= 5.16
+BuildRequires:  perl-macros
+%if %{with pod}
+BuildRequires:  perl(Test::Pod)
+BuildRequires:  perl(Test::Pod::Coverage) &gt;= 1.00
+%endif
+BuildRequires:  perl(XML::DOM)
+BuildRequires:  perl(XML::XPathEngine) &gt;= 0.1
+Requires:       perl(XML::DOM)
+Requires:       perl(XML::XPathEngine) &gt;= 0.1
+
+%description
+XML::DOM::XPath allows you to use XML::XPath methods to query a DOM. This
+is often much easier than relying only on getElementsByTagName.
+
+Authors:
+--------
+    Michel Rodriguez &lt;mirod@cpan.org&gt;
+
+%prep
+%setup -q -n %{cpan_name}-%{version}
+
+%build
+%{__perl} Makefile.PL INSTALLDIRS=vendor
+%{__make} %{?_smp_mflags}
+
+%check
+%{__make} test
+
+%install
+%perl_make_install
+# do not perl_process_packlist (noarch)
+# remove .packlist file
+%{__rm} -rf $RPM_BUILD_ROOT%perl_vendorarch
+# remove perllocal.pod file
+%{__rm} -rf $RPM_BUILD_ROOT%perl_archlib
+%perl_gen_filelist
+
+%clean
+%{__rm} -rf $RPM_BUILD_ROOT
+
+%files -f %{name}.files
+%defattr(-,root,root,-)
+%doc Changes README
+
+%changelog
</diff>
    </file>
  </files>
</sourcediff>
</action>
</request>"""
        else:
            body = """
changes files:
--------------

++++++ new changes file:
--- perl-XML-DOM-XPath.changes
+++ perl-XML-DOM-XPath.changes
@@ -0,0 +1,22 @@
+-------------------------------------------------------------------
+Tue Aug  6 22:46:18 UTC 2013 - rcurtis@suse.com
+
+- Updated spec file to replace perl_requires macro with preferred SUSE Requires:  perl >= x.x format
+
+-------------------------------------------------------------------
+Mon Aug  5 10:10:37 UTC 2013 - cfarrell@suse.com
+
+- license update: Artistic-1.0 or GPL-1.0+
+  SPDX syntax 
+
+-------------------------------------------------------------------
+Wed Dec  1 13:36:09 UTC 2010 - coolo@novell.com
+
+- switch to perl_requires macro
+
+-------------------------------------------------------------------
+Wed Aug  4 12:51:34 UTC 2010 - chris@computersalat.de
+
+- initial package 0.14
+  * created by cpanspec 1.78
+

new:
----
  XML-DOM-XPath-0.14.tar.gz
  perl-XML-DOM-XPath.changes
  perl-XML-DOM-XPath.spec

spec files:
-----------

++++++ new spec file:
--- perl-XML-DOM-XPath.spec
+++ perl-XML-DOM-XPath.spec

        """
    elif (request.method == "POST"
          and "changereviewstate" in params.get("cmd", [])):
        status = 403
        body = "Forbidden for url: http://api.example.com/request/30902"
    else:
        status = 404
        body = """
            <status code="not_found">
              <summary>Couldn't find request with id '3090200000'
              </summary>
            </status>
        """

    headers['request-id'] = '728d329e-0e86-11e4-a748-0c84dc037c13'
    return status, headers, body


class TestRequest(OscTest):
    def setUp(self):
        super(TestRequest, self).setUp()

        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + '/request.*'),
            callback=CallbackFactory(callback)
        )
        self.mock_request(
            method=responses.POST,
            url=re.compile(self.osc.url + '/request.*'),
            callback=CallbackFactory(callback)
        )
        self.mock_request(
            method=responses.GET,
            url=re.compile(self.osc.url + r'/comments/request/\d+.*'),
            callback=CallbackFactory(callback)
        )
        self.mock_request(
            method=responses.POST,
            url=re.compile(self.osc.url + r'/comments/request/\d+.*'),
            callback=CallbackFactory(callback)
        )

    @responses.activate
    def test_get_list(self):
        response = self.osc.requests.get_list()
        self.assertEqual(response.tag, "directory")
        self.assertEqual(response.countchildren(), 12)

    @responses.activate
    def test_get(self):
        with self.subTest("no history"):
            response = self.osc.requests.get(30902)
            self.assertEqual(response.tag, "request")
            self.assertEqual(response.get("id"), "30902")
            self.assertEqual(response.get("creator"), "nemo")
            self.assertEqual(len(response.xpath("//request/history")), 0)

        with self.subTest("with history"):
            response = self.osc.requests.get(30902, withhistory=True)
            self.assertEqual(response.tag, "request")
            self.assertEqual(response.get("id"), "30902")
            self.assertEqual(response.get("creator"), "nemo")
            self.assertEqual(len(response.xpath("//request/history")), 4)

        with self.subTest("with full history"):
            response = self.osc.requests.get(30902, withfullhistory=True)
            self.assertEqual(response.tag, "request")
            self.assertEqual(response.get("id"), "30902")
            self.assertEqual(response.get("creator"), "nemo")
            self.assertEqual(len(response.xpath("//request/history")), 5)

    @responses.activate
    def test_cmd(self):
        with self.subTest("plain diff"):
            response = self.osc.requests.cmd(30902, "diff")
            self.assertTrue(isinstance(response, str))
            self.assertIn("changes files:", response)
            self.assertIn("+++ perl-XML-DOM-XPath.changes", response)
        with self.subTest("xml diff"):
            response = self.osc.requests.cmd(30902, "diff", view="xml")
            self.assertTrue(isinstance(response, ObjectifiedElement))
        with self.subTest("changereviewstate"):
            self.assertRaises(
                ValueError,
                self.osc.requests.cmd, 30902, "changereviewstate"
            )

    @responses.activate
    def test_comment(self):
        with self.subTest("get"):
            response = self.osc.requests.get_comments(30902)
            self.assertTrue(isinstance(response, ObjectifiedElement))
            self.assertEqual(len(response.findall("comment")), 2)
        with self.subTest("add"):
            response = self.osc.requests.add_comment(30902, "hello wørld")
            self.assertTrue(response)



"""
Common base classes
^^^^^^^^^^^^^^^^^^^
"""
# pylint: disable=too-few-public-methods,
import os

from lxml.etree import tounicode


class ExtensionBase:
    """
    Base class for extensions of the :py:class:`Ocs` entry point.
    """
    def __init__(self, osc_obj):
        self.osc = osc_obj


# pylint: disable=too-many-instance-attributes
class DataDir:
    """
    Compatibility layer for the ``.osc`` data directory used by the ``osc`` CLI
    """
    data_dir = ".osc"
    osclib_version_string = "1.0"

    # pylint: disable=too-many-arguments
    def __init__(self, osc, path, project, package=None, overwrite=False):
        self.osc = osc
        self.path = os.path.join(path, self.data_dir)
        self.project = project
        self.package = package
        self._apiurl = os.path.join(self.path, "_apiurl")
        self._project = os.path.join(self.path, "_project")
        self._package = os.path.join(self.path, "_package")
        self._files = os.path.join(self.path, "_files")
        self._osclib_version = os.path.join(self.path, "_osclib_version")
        if not os.path.isdir(self.path):
            os.makedirs(self.path)
            overwrite = True

        if overwrite:
            self.write_dir_contents()

    def write_dir_contents(self):
        """
        Create files with default content in ``.osc`` sub-directory
        """
        with open(self._apiurl, "w") as filehandle:
            filehandle.write(self.osc.url + os.linesep)

        with open(self._osclib_version, "w") as filehandle:
            filehandle.write(self.osclib_version_string + os.linesep)

        with open(self._project, "w") as filehandle:
            filehandle.write(self.project + os.linesep)

        if self.package:
            with open(self._package, "w") as filehandle:
                filehandle.write(self.package + os.linesep)

            with open(self._files, "w") as filehandle:
                filehandle.write(
                    tounicode(self.osc.packages.get_files(self.project,
                                                          self.package))
                )



"""
Mapping types
^^^^^^^^^^^^^

This module provides a collection of mapping types with special behavior.

.. versionadded:: 0.3.0
"""
# pylint: disable=too-many-ancestors
from collections.abc import MutableMapping


class Mappable(MutableMapping):
    """
    Basic implementation of a dictionary
    """
    def __init__(self, **kwargs):
        self._data = kwargs.copy()

    def __getitem__(self, item):
        if item not in self._data:
            self.__missing__(item)
        return self._data[item]

    def __setitem__(self, key, value):
        self._data[key] = value

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return len(self._data)

    def __delitem__(self, key):
        del self._data[key]

    def __contains__(self, item):
        return item in self._data

    def __missing__(self, key):
        raise KeyError(key)

    def __str__(self):
        return str(self._data)

    def __repr__(self):
        return repr(self._data)

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()


class LazyOscMappable(Mappable):
    """
    Base class for lazy mappables

    To define custom subclasses override the ``__missing__`` method and set the missing key.

    :param osc_obj: An Osc instance to run queries
    :type osc_obj: :py:class:`osctiny.osc.Osc`
    """
    def __init__(self, osc_obj, **kwargs):
        super().__init__(**kwargs)
        self.osc = osc_obj



"""
Search extension
----------------
"""
from __future__ import unicode_literals
from six.moves.urllib.parse import urljoin

from ..utils.base import ExtensionBase


class Search(ExtensionBase):
    """
    Osc extension providing methods for searching
    """
    base_path = "/search/"

    def search(self, path, xpath, **kwargs):
        """
        Search for objects in buildservice

        .. note:: Shortcuts

            There are shortcut method for the object types ``project``,
            ``package`` and ``request``. These only expect the ``xpath``
            argument, e.g.:

            .. code:: python

                Osc.search.project("starts-with(@name,'SUSE:Maintenance')")

        .. note:: Pagination

            Search results can be paginated using ``limit`` and ``offset``, e.g.

            .. code:: python

                Osc.search.project("starts-with(@name,'SUSE:Maintenance')",
                                   limit=10, offset=30)

        :param path: object type / relative URL
        :param xpath: XPath expression to filter results
        :param kwargs: Additional parameters to be passed to the underlying API
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        kwargs["match"] = xpath
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path + path.lstrip("/")),
            params=kwargs,
            method='GET'
        )
        return self.osc.get_objectified_xml(response)

    def __getattr__(self, name):
        allowed = ['project', 'package', 'request']
        if name not in allowed:
            raise AttributeError(
                "No such attribute: '{}'. Use one of: {}".format(
                    name, ", ".join(allowed)
                )
            )

        return lambda *args, **kwargs: self.search(name, *args, **kwargs)



# pylint: disable=missing-docstring
from .osc import Osc
from .extensions import bs_requests, buildresults, comments, packages, \
    projects, search, users


__all__ = ['Osc', 'bs_requests', 'buildresults', 'comments', 'packages',
           'projects', 'search', 'users']
__version__ = "0.6.1"



"""
Projects extension
------------------
"""
from __future__ import unicode_literals
import re
from six.moves.urllib.parse import urljoin
from six import text_type

from lxml.etree import tounicode
from lxml.objectify import fromstring

from ..utils.base import ExtensionBase


TEMPLATE_CREATE_ATTR = "<attributes><attribute namespace='' name=''>" \
                       "<value></value></attribute></attributes>"
TEMPLATE_META = "<project name=''><title></title><description></description>" \
                "<person userid='' role='bugowner'/>" \
                "<person userid='' role='maintainer'/>" \
                "<build><enable/></build><publish><disable/></publish>" \
                "<debuginfo><enable/></debuginfo></project>"


class Project(ExtensionBase):
    """
    Osc extension to interact with projects
    """
    base_path = "/source"
    attribute_pattern = re.compile(r"^((?P<prefix>[^:]+):)?(?P<name>.+)$")

    def get_list(self, deleted=False):
        """
        Get list of projects

        :param deleted: show deleted projects instead of existing
        :type deleted: bool
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path),
            method="GET",
            params={'deleted': deleted}
        )

        return self.osc.get_objectified_xml(response)

    def get_meta(self, project):
        """
        Get project metadata

        :param project: name of project
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "{}/{}/_meta".format(self.base_path, project)
            ),
            method="GET"
        )

        return self.osc.get_objectified_xml(response)

    # pylint: disable=too-many-arguments
    def put_meta(self, project, metafile=None, title=None, description=None,
                 bugowner=None, maintainer=None):
        """
        Edit project meta data or create a new project

        If no ``metafile`` is provided, a default template is used.

        .. versionadded:: 0.1.5

        :param project: name of project
        :param metafile: Complete metafile
        :type metafile: str or ElementTree
        :param title: Title for meta file
        :param description: Description for meta file
        :param bugowner: Bugowner for meta file
        :param maintainer: Maintainer for meta file
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        if metafile is None:
            metafile = TEMPLATE_META

        if isinstance(metafile, text_type):
            metafile = fromstring(metafile)

        metafile.set("name", project)

        # pylint: disable=protected-access
        if title:
            metafile.title._setText(title)
        if description:
            metafile.description._setText(description)
        if bugowner:
            person = metafile.xpath("person[@role='bugowner']")
            if person:
                person[0].set("userid", bugowner)
        if maintainer:
            person = metafile.xpath("person[@role='maintainer']")
            if person:
                person[0].set("userid", maintainer)

        response = self.osc.request(
            url=urljoin(self.osc.url,
                        "/".join((self.base_path, project, "_meta"))),
            method="PUT",
            data=tounicode(metafile)
        )

        parsed = self.osc.get_objectified_xml(response)
        if response.status_code == 200 and parsed.get("code") == "ok":
            return True

        return parsed

    def get_files(self, project, directory="", meta=False, rev=None, **kwargs):
        """
        List project files

        :param project: name of project
        :param directory: directory in project
        :param meta: switch for _meta files
        :type meta: bool
        :param rev: revision
        :type rev: int
        :param kwargs: More keyword arguments for API call
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        if meta:
            kwargs["meta"] = '1'
        if rev:
            kwargs["rev"] = text_type(rev)
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "{}/{}/{}".format(self.base_path, project, directory)
            ),
            method="GET",
            params=kwargs
        )

        return self.osc.get_objectified_xml(response)

    def get_attribute(self, project, attribute=None):
        """
        Get one attribute of a project

        .. note::

            Be aware of namespace prefixes.

            When specifying the ``attribute`` argument make sure to include the
            namespace prefix and separate both by a colon, e.g.
            ``OBS:IncidentPriority``.

        :param project: name of project
        :param attribute: name of attribute
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        url = urljoin(
            self.osc.url,
            "{}/{}/_attribute".format(
                self.base_path, project
            )
        )

        if attribute:
            url = "{}/{}".format(url, attribute)

        response = self.osc.request(url=url, method="GET")

        return self.osc.get_objectified_xml(response)

    def set_attribute(self, project, attribute, value):
        """
        Set or update an attribute of a project

        :param project: project name
        :param attribute: attribute name (can include prefix separated by colon)
        :param value: attribute value
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        url = urljoin(
            self.osc.url,
            "{}/{}/_attribute".format(
                self.base_path, project
            )
        )
        match = self.attribute_pattern.match(attribute)
        if match is None:
            raise ValueError("Invalid attribute format: {}".format(attribute))

        attr_xml = fromstring(TEMPLATE_CREATE_ATTR)
        attr_xml.attribute.set('namespace', match.group("prefix"))
        attr_xml.attribute.set('name', match.group("name"))
        # pylint: disable=protected-access
        attr_xml.attribute.value._setText(text_type(value))

        response = self.osc.request(
            url=url,
            method="POST",
            data=tounicode(attr_xml)
        )

        parsed = self.osc.get_objectified_xml(response)
        if response.status_code == 200 and parsed.get("code") == "ok":
            return True

        return parsed

    def delete_attribute(self, project, attribute):
        """
        Delete an attribute of a project

        :param project: name of project
        :param attribute: name of attribute
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        url = urljoin(
            self.osc.url,
            "{}/{}/_attribute/{}".format(
                self.base_path, project, attribute
            )
        )

        response = self.osc.request(
            url=url,
            method="DELETE",
        )
        parsed = self.osc.get_objectified_xml(response)
        if response.status_code == 200 and parsed.get("code") == "ok":
            return True

        return parsed

    def get_comments(self, project):
        """
        Get a list of comments for project

        .. versionchanged:: 0.1.8
            Use internally :py:class:`osctiny.comments.Comment.get`

        :param project: name of project
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        return self.osc.comments.get(
            obj_type="project",
            ids=(project,)
        )

    def add_comment(self, project, comment, parent_id=None):
        """
        Add a comment to a project

        .. versionadded: 0.1.2

        .. versionchanged:: 0.1.8
            Use internally :py:class:`osctiny.comments.Comment.add`

        :param project: name of project
        :param comment: Comment to be added
        :param parent_id: ID of parent comment. Default: ``None``
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        return self.osc.comments.add(
            obj_type="project",
            ids=(project,),
            comment=comment,
            parent_id=parent_id
        )

    def get_history(self, project, meta=True, rev=None, **kwargs):
        """
        Get history of project

        To get just a particular revision, use the ``rev`` argument.

        :param project: name of project
        :param meta: Switch between meta and non-meta (normally empty) revision
                     history
        :type meta: bool
        :param rev: History revision ID
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        if rev:
            kwargs["rev"] = rev
        kwargs["meta"] = "1" if meta else "0"

        response = self.osc.request(
            url=urljoin(self.osc.url, "{}/{}/_project/_history".format(
                self.base_path, project
            )),
            method="GET",
            params=kwargs
        )

        return self.osc.get_objectified_xml(response)

    def delete(self, project, force=False, comment=None):
        """
        Delete project

        .. versionadded:: 0.1.2

        :param project: Project name
        :param force: Delete project even if subprojects, packages or pending
                      requests for those packages exist
        :param comment: Optional comment
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        params = {'force': force}

        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "/".join((self.base_path, project))
            ),
            method="DELETE",
            params=params,
            data=comment
        )

        parsed = self.osc.get_objectified_xml(response)
        if response.status_code == 200 and parsed.get("code") == "ok":
            return True

        return parsed

    def exists(self, project):
        """
        Check whether project exists

        .. versionadded:: 0.1.2

        :param project: Project name
        :return: ``True``, if package exists, otherwise ``False``
        """
        response = self.osc.request(
            url=urljoin(
                self.osc.url,
                "/".join((self.base_path, project))
            ),
            method="HEAD",
            raise_for_status=False
        )

        return response.status_code == 200

    create = put_meta



"""
Authentication handlers for 2FA
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. versionadded:: 0.6.0
"""
import typing
from base64 import b64decode, b64encode
from pathlib import Path
from subprocess import Popen, PIPE
import re
import sys
from time import time

from requests.auth import HTTPDigestAuth
from requests.cookies import extract_cookies_to_jar
from requests.utils import parse_dict_header
from requests import Response

from .errors import OscError


class HttpSignatureAuth(HTTPDigestAuth):
    """
    Implementation of the "Signature authentication scheme"

    .. note::

        This seems to be a variation of the `HTTP Message Signatures`_ specification.

        See also the `reference implementation for osc`_

        .. _HTTP Message Signatures:
            https://datatracker.ietf.org/doc/draft-ietf-httpbis-message-signatures/

        .. _reference implementation for osc: https://github.com/openSUSE/osc/pull/1032

    :param username: The username
    :param password: Passphrase for SSH key. This can be omitted, if ``ssh-agent`` is also installed
    :param ssh_key_file: Path of SSK key
    """
    def __init__(self, username: str, password: typing.Optional[str], ssh_key_file: Path):
        super().__init__(username=username, password=password)
        if not ssh_key_file.is_file():
            raise FileNotFoundError(f"SSH key at location does not exist: {ssh_key_file}")
        self.ssh_key_file = ssh_key_file
        self.pattern = re.compile(r"(?<=\)) (?=\()")

    def __eq__(self, other: 'HttpSignatureAuth') -> bool:
        return self.ssh_key_file == getattr(other, 'ssh_key_file', None) and super().__eq__(other)

    def split_headers(self, headers: str) -> typing.List[str]:
        """
        Split ``headers`` parameter from ``WWW-Authenticate Signature`` header

        :param headers: Value of the ``headers`` parameter
        """
        parts = self.pattern.split(headers)
        return [part.strip("()") for part in parts]

    def ssh_sign(self) -> str:
        """
        Solve the challenge via SSH signing
        """
        data = "\n".join(f"({header}): {self._thread_local.chal[header]}"
                         for header in self._thread_local.chal["headers"])
        cmd = ['ssh-keygen', '-Y', 'sign', '-f', self.ssh_key_file.as_posix(), '-q',
               '-n', self._thread_local.chal.get('realm', '')]
        if self.password:
            cmd += ['-P', self.password]

        encoding = sys.getdefaultencoding()
        with Popen(cmd, stdout=PIPE, stderr=PIPE, stdin=PIPE) as proc:
            signature, error = proc.communicate(data.encode(encoding))
            if proc.returncode:
                raise OscError(f"ssh-keygen returned {proc.returncode}: {error}")

        match = re.match(br"\A-----BEGIN SSH SIGNATURE-----\n(.*)\n-----END SSH SIGNATURE-----",
                         signature, re.S)
        if not match:
            raise OscError("Could not generate challenge response")
        return b64encode(b64decode(match.group(1))).decode(encoding)

    def build_digest_header(self, method: str, url: str) -> str:
        """
        Generate Authentication header
        """
        headers = " ".join(f"({header})" for header in self._thread_local.chal["headers"])
        return f'Signature keyId="{self.username}",algorithm="ssh",signature={self.ssh_sign()},' \
               f'headers="{headers}",created={self._thread_local.chal["created"]}'

    def handle_401(self, r: Response, **kwargs) -> Response:
        """
        Handle authentication in case of 401

        Contents of method copied from :py:meth:`requests.auth.HTTPDigestAuth.handle_401` and edited
        """
        if not 400 <= r.status_code < 500:
            self._thread_local.num_401_calls = 1
            return r

        if self._thread_local.pos is not None:
            # Rewind the file position indicator of the body to where
            # it was to resend the request.
            r.request.body.seek(self._thread_local.pos)
        s_auth = r.headers.get('www-authenticate', '')

        if s_auth.lower().startswith("signature") and self._thread_local.num_401_calls < 2:
            self._thread_local.num_401_calls += 1

            _, challenge = s_auth.split(" ", maxsplit=1)
            challenge = parse_dict_header(challenge)
            challenge.setdefault("headers", ["created"])
            challenge["created"] = int(time())
            challenge["headers"] = self.split_headers(challenge["headers"])
            self._thread_local.chal.update(challenge)

            # The following is unchanged from :py:meth:`requests.auth.HTTPDigestAuth.handle_401`,
            # so we ignore linter issues about it.
            # pylint: disable=pointless-statement,protected-access
            r.content
            r.close()
            prep = r.request.copy()
            extract_cookies_to_jar(prep._cookies, r.request, r.raw)
            prep.prepare_cookies(prep._cookies)
            prep.headers['Authorization'] = self.build_digest_header(
                prep.method, prep.url)
            _r = r.connection.send(prep, **kwargs)
            _r.history.append(r)
            _r.request = prep

            return _r

        self._thread_local.num_401_calls = 1
        return r



"""
Persons and groups extension
----------------------------
"""
from __future__ import unicode_literals
from six.moves.urllib.parse import urljoin

from ..utils.base import ExtensionBase


class Group(ExtensionBase):
    """
    The BuildService group API is accessible through this object.
    """
    base_path = "/group/"

    def get_list(self, login=None):
        """
        Get all groups or all groups containing user ``login``

        :param login: Username to filter for
        :rtype: lxml.objectify.ObjectifiedElement
        """
        data = {}
        if login:
            data["login"] = login
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path),
            method="GET",
            params=data
        )

        return self.osc.get_objectified_xml(response)

    def get(self, title):
        """
        Get details for group ``title``

        :param title: Title/name of the group
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path + title),
            method="GET",
        )

        return self.osc.get_objectified_xml(response)


# pylint: disable=too-few-public-methods
class Person(ExtensionBase):
    """
    The BuildService person API is accessible through this object.
    """
    base_path = "/person/"

    def get(self, userid):
        """
        Get details for user ``userid``

        :param userid: Username
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path + userid),
            method="GET",
        )

        return self.osc.get_objectified_xml(response)



"""
Base classes for osc-tiny specific exceptions
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
"""


class OscError(Exception):
    """
    Base class for expcetions to be raised by ``osctiny``
    """



"""
Issues extension
----------------
"""
from __future__ import unicode_literals
import os

from six.moves.urllib.parse import urljoin
from six import text_type

from ..utils.backports import lru_cache
from ..utils.base import ExtensionBase


class Issue(ExtensionBase):
    """
    The BuildService issue(-tracker) API is accessible through this object.

    .. versionadded:: 0.2.2
    """
    base_path = "/issue_trackers/"

    @staticmethod
    def _validate(value):
        return text_type(value)

    @lru_cache(maxsize=1)
    def get_trackers(self):
        """
        Get all issue trackers with data
        """
        response = self.osc.request(
            url=urljoin(self.osc.url, self.base_path),
            method="GET",
        )
        return self.osc.get_objectified_xml(response)

    def get_tracker(self, name):
        """
        Get information on one issue tracker

        :param str name: issue tracker name
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        response = self.osc.request(
            url=urljoin(self.osc.url, os.path.join(self.base_path, name)),
            method="GET",
        )
        return self.osc.get_objectified_xml(response)

    def get(self, tracker, name, force_update=None):
        """
        Get details for an issue

        :param str tracker: issue tracker name
        :param str name: issue name
        :param force_update: If ``True``, BuildService will update the issue
                             details internally prior to returning the response
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        tracker, name = [self._validate(x) for x in [tracker, name]]
        trackers = self.get_trackers()
        can_get_details = trackers.xpath(
            "issue-tracker/name[text()='{}']/../enable-fetch".format(tracker))

        response = self.osc.request(
            url=urljoin(self.osc.url,
                        os.path.join(self.base_path, tracker, "issues", name)),
            method="GET",
            params={'force_update': force_update}
        )
        response = self.osc.get_objectified_xml(response)

        if not force_update and can_get_details and all(can_get_details):
            if not getattr(getattr(response, "summary", None), "text", None):
                return self.get(tracker, name, force_update=True)

        return response



"""
Comments extension
------------------
"""
from __future__ import unicode_literals
import os
from six.moves.urllib.parse import urljoin
from six import text_type

from ..utils.base import ExtensionBase


class Comment(ExtensionBase):
    """
    The BuildService comment(s) API is accessible through this object.
    """
    base_path = "/comments/"
    allowed_object_types = {"project": 1, "package": 2, "request": 1}

    def _validate(self, obj_type, ids):
        if obj_type not in self.allowed_object_types:
            raise ValueError(
                "Type '{}' is not in list of allowed values: {}".format(
                    obj_type, self.allowed_object_types
                )
            )

        required_length = self.allowed_object_types.get(obj_type, -1)
        if len(ids) != required_length:
            raise ValueError("For object type {} the length of IDs has to be "
                             "equal to {}!".format(obj_type, required_length))

    def get(self, obj_type, ids):
        """
        Get a list of comments for object

        .. versionadded:: 0.1.8

        .. note::

            The ``ids`` argument will look like this, e.g. for a package:
            ``('project_name', 'package_name')``.


        :param obj_type: Name of object type (project, package, request)
        :type obj_type: str
        :param ids: IDs of object
        :type ids: iterable
        :return: Objectified XML element
        :rtype: lxml.objectify.ObjectifiedElement
        """
        self._validate(obj_type, ids)
        response = self.osc.request(
            url=urljoin(self.osc.url,
                        os.path.join(*([self.base_path, obj_type]
                                       + [text_type(x) for x in ids]))),
            method="GET",
        )
        return self.osc.get_objectified_xml(response)

    def add(self, obj_type, ids, comment, parent_id=None):
        """
        Add a comment to object

        :param obj_type: Name of object type (project, package, request)
        :type obj_type: str
        :param ids: IDs of object
        :type ids: iterable
        :param comment: Comment to be added
        :param parent_id: ID of parent comment. Default: ``None``
        :return: ``True``, if successful.
        :raises HTTPError: if comment was not saved correctly. The raised exception contains the
                           full response object and API response.

        .. versionadded: 0.1.8

        .. versionchanged:: 0.3.0

            Instead of simply returning the API response, if the comment was not added, an exception
            is raised.
        """
        self._validate(obj_type, ids)
        url = urljoin(self.osc.url,
                      os.path.join(*([self.base_path, obj_type] + [text_type(x) for x in ids])))
        params = {}
        if parent_id and text_type(parent_id).isnumeric():
            params["parent_id"] = parent_id

        response = self.osc.request(
            url=url,
            method="POST",
            data=comment,
            params=params,
            raise_for_status=True
        )
        parsed = self.osc.get_objectified_xml(response)
        if response.status_code == 200 and parsed.get("code") == "ok":
            return True

        return False

    def delete(self, comment_id):
        """
        Delete comment

        .. versionadded:: 0.1.8

        :param comment_id: ID of comment
        :return: ``True``, if successful. Otherwise API response
        :rtype: bool or lxml.objectify.ObjectifiedElement
        """
        url = urljoin(self.osc.url, '/comment/' + text_type(comment_id))

        response = self.osc.request(
            url=url,
            method="DELETE",
            raise_for_status=False
        )

        parsed = self.osc.get_objectified_xml(response)
        if response.status_code == 200 and parsed.get("code") == "ok":
            return True

        return parsed



# -*- coding: utf-8 -*-
"""
Changelog
^^^^^^^^^

Parser and generator for SUSE/OpenBuildService style changelog files.

This parser/generator is a reference implementation of the ``vc`` bash script
from the openSUSE:Tools ``build`` package. See:
`<https://build.opensuse.org/package/show/openSUSE:Tools/build>`_

.. versionadded:: 0.1.11
"""
from datetime import datetime
from io import TextIOBase
import re
import warnings

from dateutil.parser import parse
from pytz import _UTC


def is_aware(timestamp):
    """
    Check whether timestamp is timezone aware

    :param timestamp: :py:class:`datetime.datetime`
    :return: ``True``, if timestamp is tz-aware
    """
    return timestamp.tzinfo is not None \
           and timestamp.tzinfo.utcoffset(timestamp) is not None


class Entry:
    """
    Representation of a complete changelog entry

    .. py:attribute:: timestamp

        Timestamp of the entry as a :py:class:`datetime.datetime` object. If no
        value is provided during initialization,
        :py:func:`datetime.datetime.now` is used instead.

    .. py:attribute:: packager

        Email and name of the packager. Valid formats are:

        * ``email@example.com``
        * ``<email@example.com>``
        * ``full name <email@example.com>``

    .. py:attribute:: content

        All lines until the beginning of the next entry; except empty lines at
        the beginning and end
    """
    timestamp = None
    packager = None
    content = ""
    default_tz = _UTC()

    def __init__(self, timestamp=None, packager=None, content=""):
        if not isinstance(timestamp, datetime) and timestamp is not None:
            raise TypeError("`timestamp` needs to be a datetime object!")
        if timestamp and not is_aware(timestamp):
            raise ValueError("`timestamp` is not timezone-aware!")
        self.timestamp = timestamp or self.now()
        self.packager = packager
        self.content = content

    def __bool__(self):
        return bool(self.timestamp and self.packager and self.content)

    def __len__(self):
        return 1 if self.timestamp and self.packager and self.content else 0

    def now(self):
        """
        Return current UTC timestamp

        :return: :py:class:`datetime.datetime`
        """
        return datetime.now(tz=self.default_tz)

    @property
    def formatted_timestamp(self):
        """
        Return properly formatted timestamp

        :return: str
        """
        if not isinstance(self.timestamp, datetime):
            return self.timestamp

        return self.timestamp\
            .astimezone(self.default_tz)\
            .strftime("%a %b %d %H:%M:%S %Z %Y")

    def __str__(self):
        return "{sep}\n{self.formatted_timestamp} - {self.packager}\n\n" \
               "{self.content}\n\n".format(sep="-" * 67, self=self)

    def __unicode__(self):
        return self.__str__()


class ChangeLog:
    """
    Provider of capabilities to parse and write a ``.changes`` file

    Parsing and starting a new ``.changes`` file should be intuitive. In order
    to append or edit entries use this recipe:

    .. code:: python

        cl = ChangeLog.parse(path="/path/to/file", generative=False)
        cl.entries.append(Entry(...))
        cl.write()

    .. py:attribute:: entry_factory

        A class used to store entry data.

        **Example**: :py:class:`Entry` could be extended to provide a read-only
        property to find bug references inside the entry's content.
    """
    additional_tzinfos = {}
    entry_factory = Entry
    entries = None
    patterns = {
        "init": re.compile(r"^-+$"),
        "header": re.compile(
            r"^(?P<timestamp>[A-Za-z]{3,} [A-Za-z]{3,} [0-9: ]+ [A-Z]+ \d{4,}) "
            r"- (?P<packager>.*)\s*$")
    }

    def __init__(self):
        self.entries = []

    def _parse(self, handle):
        """
        Actual method for parsing.

        .. important::

            Please do not call this method directly. Use :py:meth:`parse`
            instead.

        :param handle: An open and iterable (file) handle
        :type handle: Any derived object of :py:class:`io.IOBase`
        """
        # pylint: disable=too-many-branches,consider-using-with
        entry = self.entry_factory()

        if isinstance(handle, TextIOBase):
            handle.seek(0)
        elif isinstance(handle, (str, bytes)):
            handle = open(handle, "r")  # pylint: disable=consider-using-with
        else:
            raise TypeError("Unexpected type for 'path': {}".format(
                type(handle)))

        try:
            for line in handle:
                match = self.patterns["init"].match(line)
                if match:
                    if entry:
                        # We are at the beginning of a new entry. Time to emit
                        # the finished one.
                        yield entry
                    entry = self.entry_factory()
                    continue

                match = self.patterns["header"].match(line)
                if match:
                    try:
                        entry.timestamp = parse(match.group("timestamp"),
                                                ignoretz=False,
                                                tzinfos=self.additional_tzinfos)
                    except ValueError as error:
                        warnings.warn(
                            "Cannot parse changelog entry's timestamp: "
                            "'{}'".format(error))
                        entry.timestamp = match.group("timestamp")
                    else:
                        # Assuming UTC may not be correct, but beats dealing
                        # with a mix of tz-aware and naive datetime objects
                        if not is_aware(entry.timestamp):
                            entry.timestamp = entry.default_tz.localize(
                                parse(match.group("timestamp"), ignoretz=True,)
                            )
                    entry.packager = match.group("packager")
                    continue

                if not line.strip():
                    continue

                if entry.content:
                    entry.content += "\n"

                entry.content += line.rstrip()

            if entry:
                # The last entry of the file is emitted explicitly
                yield entry
        finally:
            handle.close()

    @classmethod
    def parse(cls, path, generative=True):
        """
        Parse a changes file

        The changes file to parse may be specified by it's path or as an already
        open file handle (aka. subclass of :py:class:`io.TextIOBase`).

        Use this method to initialize a new instance of :py:class:`ChangeLog`
        like this:

        .. code:: python

            cl = ChangeLog.parse(path="/path/to/file")
            for entry in cl.entries:
                print(entry)

        :param path: If a path is supplied as a string, the file will be opened
                     for reading. If a subclass of :py:class:`io.TextIOBase` is
                     provided, this method assumes that it is opened for
                     reading.
        :type path: str or open file/stream handle
        :param bool generative: If set to ``True`` (default), changelog entries
                                are parsed on the fly using a generator.
                                Otherwise all entries are parsed immediately and
                                stored in an iterable.
        :return: An instance of ``ChangeLog``
        :raises TypeError: if ``path`` is not a string or a subclass of
                           :py:class:`io.TextIOBase`
        """
        new = cls()

        # pylint: disable=protected-access
        if generative:
            new.entries = new._parse(path)
        else:
            new.entries = list(new._parse(path))

        return new

    def write(self, path):
        """
        Write entries to file/stream

        This method can write to files and any other stream derived from
        :py:class:`io.IOBase`.

        :param path: Path to file or open, writable handle
        :type path: str or  :py:class:`io.IOBase`
        """
        def _wrapped(handle):
            for entry in sorted(self.entries, key=lambda x: x.timestamp,
                                reverse=True):
                handle.write(str(entry))

        if isinstance(path, TextIOBase):
            _wrapped(path)
        elif isinstance(path, (str, bytes)):
            with open(path, "w") as handle:
                _wrapped(handle)
        else:
            raise TypeError("Unexpected type for 'path': {}".format(type(path)))



import re

from lxml.objectify import ObjectifiedElement
from requests.exceptions import HTTPError
import responses

from .base import OscTest, CallbackFactory


class TestComment(OscTest):
    @responses.activate
    def test_delete_comment(self):
        def callback(headers, params, request):
            comment_id = -1
            status, body = 404, """
            <status code="not_found">
                <summary>Couldn't find Comment with 'id'=******</summary>
            </status>"""

            match = re.search(r"/(?P<id>\d+/?$)", request.url)
            if match:
                comment_id = int(match.group("id"))

            if comment_id == 666:
                status, body = 200, """<status code="ok"><summary/></status>"""

            return status, headers, body

        self.mock_request(
            method=responses.DELETE,
            url=re.compile(self.osc.url + r'/comment/\d+'),
            callback=CallbackFactory(callback)
        )

        with self.subTest("Existing comment"):
            self.assertTrue(self.osc.comments.delete(666))

        with self.subTest("Non-existing comment"):
            response = self.osc.comments.delete(1)
            self.assertTrue(isinstance(response, ObjectifiedElement))

    @responses.activate
    def test_add_comment(self):
        def callback(headers, params, request):
            status, body = 400, """
            <status code="invalid_record">
              <summary>Body is too long (maximum is 6 characters)</summary>
            </status>
            """
            if len(request.body) <= 6:
                status = 200
                body = '<status code="ok"/>'
            return status, headers, body

        self.mock_request(
            method=responses.POST,
            url=re.compile(self.osc.url + r'/comments/project/.+'),
            callback=CallbackFactory(callback)
        )

        with self.subTest("Good comment"):
            self.assertTrue(self.osc.comments.add("project", ("home:nemo",), "foo"))

        with self.subTest("Bad comment"):
            self.assertRaises(HTTPError, self.osc.comments.add, 'project', ('home:nemo',),
                              'don\'t panic')
